# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import functools
import math
import os
import re
from typing import Any, Dict, List, Optional, Union

from ... import __version__
from ..._mars.utils import calc_size_by_str, parse_readable_size
from ...compat._constants import COMPATIBLE_DEPS
from ...utils import get_local_package_version, get_local_py_version
from ._constants import SERVICE_PID_FILE

try:
    from kubernetes.client import ApiClient
except ImportError:  # pragma: no cover
    ApiClient = None

DEFAULT_WORKER_CACHE_MEM = "40%"


def _remove_nones(cfg) -> Dict:
    return dict((k, v) for k, v in cfg.items() if v is not None)


_kube_api_mapping = {
    "v1": "CoreV1Api",
    "apps/v1": "AppsV1Api",
    "rbac.authorization.k8s.io/v1": "RbacAuthorizationV1Api",
}


@functools.lru_cache(10)
def _get_k8s_api(api_version: str, k8s_api_client: ApiClient):
    from kubernetes import client as kube_client

    return getattr(kube_client, _kube_api_mapping[api_version])(k8s_api_client)


@functools.lru_cache(10)
def _camel_to_underline(name: str) -> str:
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


class KubeConfig(abc.ABC):
    api_version = "v1"

    def create_namespaced(self, k8s_api_client: ApiClient, namespace: str):
        api = _get_k8s_api(self.api_version, k8s_api_client)
        config = self.build()
        method_name = f'create_namespaced_{_camel_to_underline(config["kind"])}'
        return getattr(api, method_name)(namespace, config)

    @abc.abstractmethod
    def build(self):
        """Build config dict of the object"""


class RoleConfig(KubeConfig):
    """
    Configuration builder for Kubernetes RBAC roles
    """

    api_version = "rbac.authorization.k8s.io/v1"

    def __init__(
        self, name: str, namespace: str, api_groups: str, resources: str, verbs: str
    ):
        self._name = name
        self._namespace = namespace
        self._api_groups = api_groups.split(",")
        self._resources = resources.split(",")
        self._verbs = verbs.split(",")

    def build(self):
        return {
            "kind": "Role",
            "metadata": {"name": self._name, "namespace": self._namespace},
            "rules": [
                {
                    "apiGroups": self._api_groups,
                    "resources": self._resources,
                    "verbs": self._verbs,
                },
                {
                    "apiGroups": ["apps"],
                    "resources": ["deployments", "deployments/scale"],
                    "verbs": ["get", "watch", "list", "patch"],
                },
            ],
        }


class RoleBindingConfig(KubeConfig):
    """
    Configuration builder for Kubernetes RBAC role bindings
    """

    api_version = "rbac.authorization.k8s.io/v1"

    def __init__(
        self, name: str, namespace: str, role_name: str, service_account_name: str
    ):
        self._name = name
        self._namespace = namespace
        self._role_name = role_name
        self._service_account_name = service_account_name

    def build(self):
        return {
            "kind": "RoleBinding",
            "metadata": {"name": self._name, "namespace": self._namespace},
            "roleRef": {
                "apiGroup": "rbac.authorization.k8s.io",
                "kind": "Role",
                "name": self._role_name,
            },
            "subjects": [
                {
                    "kind": "ServiceAccount",
                    "name": self._service_account_name,
                    "namespace": self._namespace,
                }
            ],
        }


class NamespaceConfig(KubeConfig):
    """
    Configuration builder for Kubernetes namespaces
    """

    def __init__(self, name: str):
        self._name = name

    def build(self):
        return {
            "kind": "Namespace",
            "metadata": {
                "name": self._name,
                "labels": {
                    "name": self._name,
                },
            },
        }


class ServiceConfig(KubeConfig):
    """
    Configuration builder for Kubernetes services
    """

    def __init__(
        self,
        name: str,
        service_type: Optional[str],
        selector: Union[str, Dict],
        port: int,
        target_port: Optional[int] = None,
        protocol: Optional[str] = None,
    ):
        self._name = name
        self._type = service_type
        self._protocol = protocol or "TCP"
        self._selector = selector
        self._port = port
        self._target_port = target_port

    def build(self):
        return {
            "kind": "Service",
            "metadata": {
                "name": self._name,
                "labels": {
                    "xorbits/service-name": self._name,
                },
            },
            "spec": _remove_nones(
                {
                    "type": self._type,
                    "selector": self._selector,
                    "ports": [
                        _remove_nones(
                            {
                                "protocol": self._protocol,
                                "port": self._port,
                                "targetPort": self._target_port,
                            }
                        ),
                    ],
                }
            ),
        }


class IngressConfig(KubeConfig):
    api_version = "networking.k8s.io/v1"

    def __init__(
        self,
        namespace: str,
        name: str,
        service_name: str,
        service_port: int,
        cluster_type: str,
    ):
        self._namespace = namespace
        self._name = name
        self._service_name = service_name
        self._service_port = service_port
        self._cluster_type = cluster_type

    def build(self):
        from kubernetes import client

        annotations = None
        ingress_cls_name = None
        if self._cluster_type == "eks":
            annotations = {
                "alb.ingress.kubernetes.io/scheme": "internet-facing",
                "alb.ingress.kubernetes.io/target-type": "ip",
            }
            ingress_cls_name = "alb"

        body = client.V1Ingress(
            api_version="networking.k8s.io/v1",
            kind="Ingress",
            metadata=client.V1ObjectMeta(
                name=self._name, namespace=self._namespace, annotations=annotations
            ),
            spec=client.V1IngressSpec(
                rules=[
                    client.V1IngressRule(
                        http=client.V1HTTPIngressRuleValue(
                            paths=[
                                client.V1HTTPIngressPath(
                                    path="/",
                                    path_type="Prefix",
                                    backend=client.V1IngressBackend(
                                        service=client.V1IngressServiceBackend(
                                            name=self._service_name,
                                            port=client.V1ServiceBackendPort(
                                                number=self._service_port
                                            ),
                                        )
                                    ),
                                )
                            ]
                        )
                    )
                ],
                ingress_class_name=ingress_cls_name,
            ),
        )

        return body


class ResourceConfig:
    """
    Configuration builder for Kubernetes computation resources
    """

    def __init__(self, cpu, memory):
        self._cpu = cpu
        self._memory, ratio = (
            parse_readable_size(memory) if memory is not None else (None, False)
        )
        assert not ratio

    @property
    def cpu(self) -> int:
        return self._cpu

    @property
    def memory(self) -> float:
        return self._memory

    def build(self):
        return _remove_nones(
            {
                "cpu": f"{int(self._cpu * 1000)}m" if self._cpu else None,
                "memory": str(int(self._memory)) if self._memory else None,
            }
        )


class PortConfig:
    """
    Configuration builder for Kubernetes ports definition for containers
    """

    def __init__(self, container_port: Union[int, str]):
        self._container_port = int(container_port)

    def build(self):
        return {
            "containerPort": self._container_port,
        }


class VolumeConfig(abc.ABC):
    """
    Base configuration builder for Kubernetes volumes
    """

    def __init__(self, name: str, mount_path: str):
        self.name = name
        self.mount_path = mount_path

    @abc.abstractmethod
    def build(self):
        """Build volume config"""

    def build_mount(self):
        return {
            "name": self.name,
            "mountPath": self.mount_path,
        }


class HostPathVolumeConfig(VolumeConfig):
    """
    Configuration builder for Kubernetes host volumes
    """

    def __init__(
        self,
        name: str,
        mount_path: str,
        host_path: str,
        volume_type: Optional[str] = None,
    ):
        super().__init__(name, mount_path)
        self._host_path = host_path
        self._volume_type = volume_type or "DirectoryOrCreate"

    def build(self):
        return {
            "name": self.name,
            "hostPath": {"path": self._host_path, "type": self._volume_type},
        }


class EmptyDirVolumeConfig(VolumeConfig):
    """
    Configuration builder for Kubernetes empty-dir volumes
    """

    def __init__(
        self,
        name: str,
        mount_path: str,
        use_memory: bool = True,
        size_limit: Optional[int] = None,
    ):
        super().__init__(name, mount_path)
        self._medium = "Memory" if use_memory else None
        self._size_limit = size_limit

    def build(self):
        result = {"name": self.name, "emptyDir": {}}
        if self._medium:
            result["emptyDir"]["medium"] = self._medium
        if self._size_limit:
            result["emptyDir"]["sizeLimit"] = str(int(self._size_limit))
        return result


class ContainerEnvConfig:
    """
    Configuration builder for Kubernetes container environments
    """

    def __init__(
        self, name: str, value: Optional[Any] = None, field_path: Optional[str] = None
    ):
        self._name = name
        self._value = value
        self._field_path = field_path

    def build(self):
        result = dict(name=self._name)
        if self._value is not None:
            result["value"] = str(self._value)
        elif self._field_path is not None:  # pragma: no branch
            result["valueFrom"] = {"fieldRef": {"fieldPath": self._field_path}}
        return result


class ProbeConfig:
    """
    Base configuration builder for Kubernetes liveness and readiness probes
    """

    def __init__(
        self,
        initial_delay: int = 5,
        period: int = 5,
        timeout: Optional[int] = None,
        success_thresh: Optional[int] = None,
        failure_thresh: Optional[int] = None,
    ):
        self._initial_delay = initial_delay
        self._period = period
        self._timeout = timeout
        self._success_thresh = success_thresh
        self._failure_thresh = failure_thresh

    def build(self):
        return _remove_nones(
            {
                "initialDelaySeconds": self._initial_delay,
                "periodSeconds": self._period,
                "timeoutSeconds": self._timeout,
                "successThreshold": self._success_thresh,
                "failureThreshold": self._failure_thresh,
            }
        )


class TcpSocketProbeConfig(ProbeConfig):
    """
    Configuration builder for TCP liveness and readiness probes
    """

    def __init__(self, port: int, **kwargs):
        super().__init__(**kwargs)
        self._port = port

    def build(self):
        ret = super().build()
        ret["tcpSocket"] = {"port": self._port}
        return ret


class CommandExecProbeConfig(ProbeConfig):
    """
    Configuration builder for command probe
    """

    def __init__(self, commands: List[str], **kwargs):
        super().__init__(**kwargs)
        self._commands = commands

    def build(self):
        ret = super().build()
        ret["exec"] = {"command": self._commands}
        return ret


class ReplicationConfig(KubeConfig):
    """
    Base configuration builder for Kubernetes replication controllers
    """

    _default_kind = "Deployment"

    def __init__(
        self,
        name: Optional[str],
        image: str,
        replicas: int,
        resource_request: Optional["ResourceConfig"] = None,
        resource_limit: Optional["ResourceConfig"] = None,
        liveness_probe: Optional["ProbeConfig"] = None,
        startup_probe: Optional["ProbeConfig"] = None,
        pre_stop_command: Optional[List[str]] = None,
        kind: Optional[str] = None,
        external_storage: Optional[str] = None,
        external_storage_config: Optional[dict] = None,
        **kwargs,
    ):
        self._name = name
        self._kind = kind or self._default_kind
        self._image = image
        self._replicas = replicas
        self._ports: List[PortConfig] = []
        self._volumes: List[VolumeConfig] = []
        self._envs: Dict[str, ContainerEnvConfig] = dict()
        self._labels: Dict[str, Any] = dict()

        self.add_default_envs()

        self._resource_request = resource_request
        self._resource_limit = resource_limit

        self._liveness_probe = liveness_probe
        self._startup_probe = startup_probe

        self._pre_stop_command = pre_stop_command
        self._external_storage = external_storage
        self._external_storage_config = external_storage_config
        self._use_local_image = kwargs.pop("use_local_image", False)
        self._pip: Optional[Union[str, List[str]]] = kwargs.pop("pip", None)
        self._conda: Optional[Union[str, List[str]]] = kwargs.pop("conda", None)

    @property
    def api_version(self):
        return "apps/v1" if self._kind in ("Deployment", "ReplicaSet") else "v1"

    def add_env(self, name, value=None, field_path=None):
        self._envs[name] = ContainerEnvConfig(name, value=value, field_path=field_path)

    def remove_env(self, name):  # pragma: no cover
        self._envs.pop(name, None)

    def add_simple_envs(self, envs):
        for k, v in envs.items() or ():
            self.add_env(k, v)

    def add_labels(self, labels):
        self._labels.update(labels)

    def add_port(self, container_port):
        self._ports.append(PortConfig(container_port))

    def add_default_envs(self):
        pass  # pragma: no cover

    def add_volume(self, vol):
        self._volumes.append(vol)

    def config_init_containers(self) -> List[Dict]:
        return []

    @staticmethod
    def get_install_content(source: Union[str, List[str]]) -> str:
        if isinstance(source, str):
            if not os.path.exists(source):
                raise ValueError(f"Cannot find the file {source} for installation.")
            with open(source, "r") as f:
                content = f.read()
        else:
            content = "\n".join(source)
        return content

    @staticmethod
    def get_compatible_packages() -> List[str]:
        deps = []
        for dep in COMPATIBLE_DEPS:
            version = get_local_package_version(dep)
            if version is not None:
                deps.append(f"{dep}=={version}")
        return deps

    @abc.abstractmethod
    def build_container_command(self) -> List:
        """Output container command"""
        cmd = ""
        # All install contents must be wrapped in double quotes to ensure the correctness
        # when passing to the shell script.
        # At the same time, each command is followed by a semicolon to separate the individual commands.
        deps = self.get_compatible_packages()
        if deps:
            # Install for consistent packages must be at the top of all commands.
            # This does not affect user behavior.
            cmd += (
                f'/srv/install.sh pip_compatible "{self.get_install_content(deps)}" ; '
            )

        if self._pip is not None:
            cmd += f'/srv/install.sh pip "{self.get_install_content(self._pip)}" ; '
        if self._conda is not None:
            if isinstance(self._conda, str):
                cmd += f'/srv/install.sh conda_yaml "{self.get_install_content(self._conda)}" ; '

            else:
                cmd += f'/srv/install.sh conda_default "{self.get_install_content(self._conda)}" ; '

        return [cmd]

    def build_container(self):
        resources_dict = {
            "requests": self._resource_request.build()
            if self._resource_request
            else None,
            "limits": self._resource_limit.build() if self._resource_limit else None,
        }
        lifecycle_dict = _remove_nones(
            {
                "preStop": {
                    "exec": {"command": self._pre_stop_command},
                }
                if self._pre_stop_command
                else None,
            }
        )
        volume_juicefs = []
        if self._external_storage == "juicefs":
            volume_juicefs.append(
                {
                    "mountPath": self._external_storage_config["mountPath"],
                    "name": "data",
                }
            )
        return _remove_nones(
            {
                "command": ["/bin/sh", "-c"],
                "args": self.build_container_command(),
                "env": [env.build() for env in self._envs.values()] or None,
                "image": self._image,
                "name": self._name,
                "resources": dict((k, v) for k, v in resources_dict.items() if v)
                or None,
                "ports": [p.build() for p in self._ports] or None,
                "volumeMounts": [vol.build_mount() for vol in self._volumes]
                + volume_juicefs,
                "livenessProbe": self._liveness_probe.build()
                if self._liveness_probe
                else None,
                "startupProbe": self._startup_probe.build()
                if self._startup_probe
                else None,
                "lifecycle": lifecycle_dict or None,
                "imagePullPolicy": "Never" if self._use_local_image else None,
            }
        )

    def build_template_spec(self) -> Dict:
        result = {
            "containers": [self.build_container()],
            "initContainers": self.config_init_containers(),
            "volumes": [vol.build() for vol in self._volumes],
        }
        if self._external_storage == "juicefs":
            external_volumes = [
                {"name": "data", "persistentVolumeClaim": {"claimName": "juicefs-pvc"}}
            ]
            result["volumes"].extend(external_volumes)

        return dict((k, v) for k, v in result.items() if v)

    def build(self):
        return {
            "kind": self._kind,
            "metadata": {
                "name": self._name,
            },
            "spec": {
                "replicas": int(self._replicas),
                "template": {
                    "metadata": {
                        "labels": _remove_nones(self._labels) or None,
                    },
                    "spec": self.build_template_spec(),
                },
            },
        }


class XorbitsReplicationConfig(ReplicationConfig, abc.ABC):
    """
    Base configuration builder for replication controllers for Xorbits
    """

    rc_name: Optional[str] = None
    default_readiness_port = 15031

    def __init__(
        self,
        replicas: int,
        cpu: Optional[int] = None,
        memory: Optional[str] = None,
        limit_resources: bool = False,
        memory_limit_ratio: Optional[float] = None,
        image: Optional[str] = None,
        modules: Optional[Union[str, List[str]]] = None,
        volumes: Optional[List[VolumeConfig]] = None,
        service_name: Optional[str] = None,
        service_port: Optional[int] = None,
        external_storage: Optional[str] = None,
        external_storage_config: Optional[dict] = None,
        **kwargs,
    ):
        self._cpu = cpu
        self._memory, ratio = (
            parse_readable_size(memory) if memory is not None else (None, False)
        )
        assert not ratio

        if isinstance(modules, str):
            self._modules: Optional[List[str]] = modules.split(",")
        else:
            self._modules = modules

        req_res = ResourceConfig(cpu, memory) if cpu or memory else None
        limit_res = (
            ResourceConfig(req_res.cpu, req_res.memory * (memory_limit_ratio or 1))
            if req_res and memory
            else None
        )

        self._service_name = service_name
        self._service_port = service_port
        self._external_storage = external_storage
        self._external_storage_config = external_storage_config

        super().__init__(
            self.rc_name,
            image or self.get_default_image(),
            replicas,
            resource_request=req_res,
            resource_limit=limit_res if limit_resources else None,
            liveness_probe=self.config_liveness_probe(),
            startup_probe=self.config_startup_probe(),
            external_storage=self._external_storage,
            external_storage_config=self._external_storage_config,
            **kwargs,
        )
        if service_port:
            self.add_port(service_port)

        for vol in volumes or ():
            self.add_volume(vol)

        self.add_labels({"xorbits/service-type": self.rc_name})

    @staticmethod
    def get_default_image():
        image = "xprobe/xorbits:v" + __version__
        image += f"-py{get_local_py_version()}"
        return image

    def add_default_envs(self):
        self.add_env("MARS_K8S_POD_NAME", field_path="metadata.name")
        self.add_env("MARS_K8S_POD_NAMESPACE", field_path="metadata.namespace")
        self.add_env("MARS_K8S_POD_IP", field_path="status.podIP")

        if self._service_name:
            self.add_env("MARS_K8S_SERVICE_NAME", str(self._service_name))
        if self._service_port:
            self.add_env("MARS_K8S_SERVICE_PORT", str(self._service_port))

        self.add_env("MARS_CONTAINER_IP", field_path="status.podIP")

        if self._cpu:
            self.add_env("MKL_NUM_THREADS", str(self._cpu))
            self.add_env("MARS_CPU_TOTAL", str(self._cpu))
            if getattr(self, "stat_type", "cgroup") == "cgroup":
                self.add_env("MARS_USE_CGROUP_STAT", "1")

        if self._memory:
            self.add_env("MARS_MEMORY_TOTAL", str(int(self._memory)))

        if self._modules:
            self.add_env("MARS_LOAD_MODULES", ",".join(self._modules))

        if self._external_storage == "juicefs":
            self.add_env("XORBITS_EXTERNAL_STORAGE", "juicefs")
            self.add_env(
                "JUICEFS_MOUNT_PATH", self._external_storage_config["mountPath"]
            )

    def config_liveness_probe(self):
        """
        Liveness probe works after the startup probe.
        If the startup probe exists, the initial_delay of the liveness probe can be smaller.
        """
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def config_startup_probe() -> "ProbeConfig":
        """
        The startup probe is used to check whether the startup is smooth.
        The initial_delay of the startup probe can be sensitive to the system.
        """
        return CommandExecProbeConfig(
            ["sh", "-c", f'until [ -f "{SERVICE_PID_FILE}" ]; do sleep 3s; done'],
            failure_thresh=5,
            initial_delay=10,
            timeout=300,
        )

    @staticmethod
    def get_local_app_module(mod_name) -> str:
        return __name__.rsplit(".", 1)[0] + "." + mod_name

    def build(self):
        result = super().build()
        if self._kind in ("Deployment", "ReplicaSet"):
            result["spec"]["selector"] = {
                "matchLabels": {"xorbits/service-type": self.rc_name}
            }
        else:
            result["spec"]["selector"] = {"xorbits/service-type": self.rc_name}
        return result


class XorbitsSupervisorsConfig(XorbitsReplicationConfig):
    """
    Configuration builder for Xorbits supervisor service
    """

    rc_name = "xorbitssupervisor"

    def __init__(self, *args, **kwargs):
        self._web_port = kwargs.pop("web_port", None)
        self._readiness_port = kwargs.pop("readiness_port", self.default_readiness_port)
        super().__init__(*args, **kwargs)
        if self._web_port:
            self.add_port(self._web_port)

    def config_liveness_probe(self) -> "TcpSocketProbeConfig":
        return TcpSocketProbeConfig(
            self._readiness_port, timeout=60, failure_thresh=10, initial_delay=0
        )

    def build_container_command(self):
        cmd = super().build_container_command()
        start_command = f"/srv/entrypoint.sh {self.get_local_app_module('supervisor')}"
        if self._service_port:
            start_command += f" -p {str(self._service_port)}"
        if self._web_port:
            start_command += f" -w {str(self._web_port)}"
        if self._cpu:
            start_command += f" --n-process {str(int(math.ceil(self._cpu)))}"
        start_command += ";"

        cmd[0] += start_command
        return cmd


class XorbitsWorkersConfig(XorbitsReplicationConfig):
    """
    Configuration builder for Xorbits worker service
    """

    rc_name = "xorbitsworker"

    def __init__(self, *args, **kwargs):
        spill_volumes = kwargs.pop("spill_volumes", None) or ()
        mount_shm = kwargs.pop("mount_shm", True)
        self._limit_resources = kwargs["limit_resources"] = kwargs.get(
            "limit_resources", True
        )
        worker_cache_mem = (
            kwargs.pop("worker_cache_mem", None) or DEFAULT_WORKER_CACHE_MEM
        )
        min_cache_mem = kwargs.pop("min_cache_mem", None)
        self._readiness_port = kwargs.pop("readiness_port", self.default_readiness_port)
        self._readiness_service_name = kwargs.pop("readiness_service_name", None)
        supervisor_web_port = kwargs.pop("supervisor_web_port", None)

        super().__init__(*args, **kwargs)

        self._spill_volumes = []
        for idx, vol in enumerate(spill_volumes):
            if isinstance(vol, str):
                path = f"/mnt/hostpath{idx}"
                self.add_volume(HostPathVolumeConfig(f"host-path-vol-{idx}", path, vol))
                self._spill_volumes.append(path)
            else:
                self.add_volume(vol)
                self._spill_volumes.append(vol.mount_path)
        if self._spill_volumes:
            self.add_env("MARS_SPILL_DIRS", ":".join(self._spill_volumes))

        if self._memory:
            size_limit = calc_size_by_str(worker_cache_mem, self._memory)
            self.add_env("MARS_CACHE_MEM_SIZE", worker_cache_mem)
        else:
            size_limit = None

        if mount_shm:
            self.add_volume(
                EmptyDirVolumeConfig(
                    "xorbits-shared", "/dev/shm", size_limit=size_limit
                )
            )

        if min_cache_mem:
            self.add_env("MARS_MIN_CACHE_MEM_SIZE", min_cache_mem)
        if supervisor_web_port:
            self.add_env("MARS_K8S_SUPERVISOR_WEB_PORT", supervisor_web_port)

    def config_liveness_probe(self) -> "TcpSocketProbeConfig":
        return TcpSocketProbeConfig(
            self._readiness_port, timeout=60, failure_thresh=10, initial_delay=0
        )

    def config_init_containers(self) -> List[Dict]:
        """
        The worker pod checks whether the readiness port of the supervisor is successfully opened in the init container.
        The worker pod will not start until the init container is executed successfully.
        This ensures that the worker must start after the supervisor.
        The image and command used by the init container refer to the Kubernetes official documentation (https://kubernetes.io/docs/concepts/workloads/pods/init-containers/).
        """
        return [
            {
                "name": "waiting-for-supervisors",
                "image": "busybox:1.28",
                "command": [
                    "sh",
                    "-c",
                    f"until nslookup {self._readiness_service_name}.$(cat /var/run/secrets/kubernetes.io/serviceaccount/namespace).svc.cluster.local; do echo waiting for supervisors; sleep 5; done",
                ],
            }
        ]

    def build_container_command(self):
        cmd = super().build_container_command()
        start_command = f"/srv/entrypoint.sh {self.get_local_app_module('worker')}"
        if self._service_port:
            start_command += f" -p {str(self._service_port)}"
        start_command += ";"

        cmd[0] += start_command
        return cmd



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .cholesky import TensorCholesky, cholesky
from .dot import TensorDot, dot
from .inner import inner, innerproduct
from .inv import TensorInv, inv
from .lu import TensorLU, lu
from .matmul import TensorMatmul, matmul
from .norm import TensorNorm, norm
from .qr import TensorQR, qr
from .randomized_svd import randomized_svd
from .solve import solve
from .solve_triangular import TensorSolveTriangular, solve_triangular
from .svd import TensorSVD, svd
from .tensordot import TensorTensorDot, tensordot
from .vdot import vdot


def _install():
    from ..core import Tensor, TensorData

    setattr(Tensor, "__matmul__", matmul)
    setattr(Tensor, "dot", dot)
    setattr(TensorData, "__matmul__", matmul)
    setattr(TensorData, "dot", dot)


_install()
del _install



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
from typing import Union

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like

from ... import opcodes
from ...core import OutputType, recursive_tile
from ...serialization.serializables import AnyField, KeyField
from ...tensor.core import TENSOR_TYPE
from ...utils import pd_release_version
from ..core import DATAFRAME_TYPE, SERIES_TYPE, DataFrame, Series
from ..initializer import DataFrame as asframe
from ..initializer import Series as asseries
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import is_index_value_identical, parse_index

# in pandas 1.0.x, __setitem__ with a list with missing items are not allowed
_allow_set_missing_list = pd_release_version[:2] >= (1, 1)


class DataFrameSetitem(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.INDEXSETVALUE

    _target = KeyField("target")
    _indexes = AnyField("indexes")
    _value = AnyField("value")

    def __init__(self, target=None, indexes=None, value=None, output_types=None, **kw):
        super().__init__(
            _target=target,
            _indexes=indexes,
            _value=value,
            _output_types=output_types,
            **kw,
        )

    @property
    def target(self):
        return self._target

    @property
    def indexes(self):
        return self._indexes

    @property
    def value(self):
        return self._value

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._target = self._inputs[0]
        if len(inputs) > 1:
            self._value = self._inputs[-1]

    @staticmethod
    def _is_scalar_tensor(t):
        return isinstance(t, TENSOR_TYPE) and t.ndim == 0

    def _call_dataframe(self, target: DataFrame, value):
        raw_target = target

        inputs = [target]
        if np.isscalar(value):
            value_dtype = np.array(value).dtype
        elif self._is_scalar_tensor(value):
            inputs.append(value)
            value_dtype = value.dtype
        else:
            if isinstance(value, (pd.Series, SERIES_TYPE)):
                value = asseries(value)
                value_dtype = value.dtype
            elif isinstance(value, (pd.DataFrame, DATAFRAME_TYPE)):
                if len(self.indexes) != value.shape[1]:  # pragma: no cover
                    raise ValueError("Columns must be same length as key")

                value = asframe(value)
                value_dtype = pd.Series(list(value.dtypes), index=self._indexes)
            elif is_list_like(value) or isinstance(value, TENSOR_TYPE):
                # convert to numpy to get actual dim and shape
                if is_list_like(value):
                    value = np.array(value)

                if value.ndim == 1:
                    value = asseries(value, index=target.index)
                    value_dtype = value.dtype
                else:
                    if len(self.indexes) != value.shape[1]:  # pragma: no cover
                        raise ValueError("Columns must be same length as key")

                    value = asframe(value, index=target.index)
                    value_dtype = pd.Series(list(value.dtypes), index=self._indexes)
            else:  # pragma: no cover
                raise TypeError(
                    "Wrong value type, could be one of scalar, Series or tensor"
                )

            if target.shape[0] == 0:
                # target empty, reindex target first
                target = target.reindex(value.index)
                inputs[0] = target
            elif value.index_value.key != target.index_value.key:
                # need reindex when target df is not empty and index different
                value = value.reindex(target.index)
            inputs.append(value)

        index_value = target.index_value
        dtypes = target.dtypes.copy(deep=True)

        try:
            dtypes.loc[self._indexes] = value_dtype
        except KeyError:
            # when some index not exist, try update one by one
            if isinstance(value_dtype, pd.Series):
                for idx in self._indexes:
                    dtypes.loc[idx] = value_dtype.loc[idx]
            else:
                for idx in self._indexes:
                    dtypes.loc[idx] = value_dtype

        columns_value = parse_index(dtypes.index, store_data=True)
        ret = self.new_dataframe(
            inputs,
            shape=(target.shape[0], len(dtypes)),
            dtypes=dtypes,
            index_value=index_value,
            columns_value=columns_value,
        )
        raw_target.data = ret.data

    def _call_series(self, target: Series, value):
        inputs = [target]
        dtype = target.dtype
        shape = target.shape
        index_value = target.index_value

        target.data = self.new_series(
            inputs, shape=shape, dtype=dtype, index_value=index_value, name=target.name
        ).data

    def __call__(self, target: Union[DataFrame, Series], value):
        if target.ndim == 2:
            self._call_dataframe(target, value)
        else:
            self._call_series(target, value)

    @classmethod
    def tile(cls, op: "DataFrameSetitem"):
        if op.target.ndim == 2:
            res = yield from cls._tile_dataframe(op)
            return res
        else:
            return cls._tile_series(op)

    @classmethod
    def _tile_series(cls, op: "DataFrameSetitem"):
        in_df = op.inputs[0]
        result_chunks = []

        for chk in in_df.chunks:
            new_op = op.copy().reset_key()
            new_op.output_types = [OutputType.series]
            params = dict(
                shape=chk.shape,
                index=chk.index,
                dtype=chk.dtype,
                index_value=chk.index_value,
            )
            result_chunks.append(new_op.new_chunk([chk], **params))

        _new_op = op.copy()
        params = op.outputs[0].params.copy()
        params["nsplits"] = in_df.nsplits
        params["chunks"] = result_chunks
        return _new_op.new_seriess(op.inputs, **params)

    @classmethod
    def _tile_dataframe(cls, op: "DataFrameSetitem"):
        from ..merge.concat import DataFrameConcat

        out = op.outputs[0]
        target = op.target
        value = op.value
        indexes = op.indexes
        columns = target.columns_value.to_pandas()
        last_column_index = target.chunk_shape[1] - 1
        is_value_scalar = np.isscalar(value) or cls._is_scalar_tensor(value)
        has_multiple_cols = getattr(out.dtypes[indexes], "ndim", 0) > 0
        target_index_to_value_index = collections.defaultdict(list)

        if has_multiple_cols:
            append_cols = [c for c in indexes if c not in columns]
        else:
            append_cols = [indexes] if indexes not in columns else []

        if not is_value_scalar:
            rechunk_arg = {}

            # check if all chunk's index_value are identical
            is_identical = is_index_value_identical(target, value)
            if not is_identical:
                # do rechunk
                if any(np.isnan(s) for s in target.nsplits[0]) or any(
                    np.isnan(s) for s in value.nsplits[0]
                ):  # pragma: no cover
                    yield

                rechunk_arg[0] = target.nsplits[0]

            if isinstance(value, DATAFRAME_TYPE):
                if len(append_cols) < len(indexes):
                    # rechunk in column dim given distribution of indexes in target chunks
                    target_col_to_chunk_index = {
                        col: head_chunk.index[1]
                        for head_chunk in target.cix[0, :]
                        for col in head_chunk.dtypes.keys()
                    }
                    value_chunk_indexes = [
                        target_col_to_chunk_index.get(vc, None) for vc in indexes
                    ]
                    col_nsplits = []
                    last_cidx = value_chunk_indexes[0]
                    match_idxes = []
                    for cidx, idx in zip(value_chunk_indexes, indexes):
                        if cidx != last_cidx:
                            target_index_to_value_index[last_cidx].append(
                                len(col_nsplits)
                            )
                            col_nsplits.append(len(match_idxes))
                            last_cidx = cidx
                            match_idxes = [idx]
                        else:
                            match_idxes.append(idx)
                    target_index_to_value_index[last_cidx].append(len(col_nsplits))
                    col_nsplits.append(len(match_idxes))

                    # merge last column indexes and keep column order
                    last_value_index = target_index_to_value_index.pop(
                        last_column_index, []
                    )
                    append_value_index = target_index_to_value_index.pop(None, [])
                    target_index_to_value_index[None] = (
                        last_value_index + append_value_index
                    )

                    rechunk_arg[1] = col_nsplits
                else:
                    target_index_to_value_index[None] = [0]
                    rechunk_arg[1] = [len(append_cols)]

            if rechunk_arg:
                value = yield from recursive_tile(value.rechunk(rechunk_arg))

        out_chunks = []
        nsplits = [list(ns) for ns in target.nsplits]
        nsplits[1][-1] += len(append_cols)
        nsplits = tuple(tuple(ns) for ns in nsplits)

        for c in target.chunks:
            result_chunk = c

            if has_multiple_cols:
                new_indexes = [vc for vc in indexes if vc in c.dtypes]
            else:
                new_indexes = [indexes] if indexes in c.dtypes else []

            if c.index[-1] == last_column_index:
                new_indexes.extend(append_cols)

            if new_indexes:
                # update needed on current chunk
                chunk_op = op.copy().reset_key()
                chunk_op._indexes = new_indexes if has_multiple_cols else new_indexes[0]

                if pd.api.types.is_scalar(value):
                    chunk_inputs = [c]
                elif is_value_scalar:
                    chunk_inputs = [c, value.chunks[0]]
                else:
                    # get proper chunk from value chunks
                    if has_multiple_cols:
                        value_chunks = []
                        target_index = (
                            None if c.index[-1] == last_column_index else c.index[1]
                        )
                        for value_index in target_index_to_value_index[target_index]:
                            value_chunk = value.cix[c.index[0], value_index]
                            value_chunks.append(value_chunk)
                        if len(value_chunks) == 1:
                            value_chunk = value_chunks[0]
                        else:
                            # concat multiple columns by order
                            shape = (
                                value_chunks[0].shape[0],
                                sum(c.shape[1] for c in value_chunks),
                            )
                            dtypes = pd.concat([c.dtypes for c in value_chunks])
                            concat_op = DataFrameConcat(output_types=op.output_types)
                            value_chunk = concat_op.new_chunk(
                                value_chunks, shape=shape, dtypes=dtypes
                            )
                    else:
                        value_chunk = value.cix[c.index[0],]

                    chunk_inputs = [c, value_chunk]

                shape = c.shape
                if append_cols and c.index[-1] == last_column_index:
                    # some columns appended at the last column of chunks
                    shape = (shape[0], shape[1] + len(append_cols))

                result_chunk = chunk_op.new_chunk(
                    chunk_inputs,
                    shape=shape,
                    index=c.index,
                )
                result_chunk._set_tileable_meta(
                    tileable_key=out.key,
                    nsplits=nsplits,
                    index_value=out.index_value,
                    columns_value=out.columns_value,
                    dtypes=out.dtypes,
                )
            out_chunks.append(result_chunk)

        params = out.params
        params["nsplits"] = nsplits
        params["chunks"] = out_chunks
        new_op = op.copy()
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def estimate_size(cls, ctx: dict, op: "DataFrameSetitem"):
        result_size = ctx[op.target.key][0]
        ctx[op.outputs[0].key] = (result_size, result_size)

    @classmethod
    def execute(cls, ctx, op: "DataFrameSetitem"):
        target = ctx[op.target.key]
        if target.ndim == 2:
            cls._execute_dataframe(ctx, op)
        else:
            cls._execute_series(ctx, op)

    @classmethod
    def _execute_series(cls, ctx, op: "DataFrameSetitem"):
        target = ctx[op.target.key]

        indexes = op.indexes
        value = op.value

        try:
            _ = target[indexes]
            indexed = True
        except KeyError:
            indexed = False

        if indexed:
            try:
                target[indexes] = value
            except ValueError:
                target = target.copy(deep=True)
                target[indexes] = value

        ctx[op.outputs[0].key] = target

    @classmethod
    def _execute_dataframe(cls, ctx, op: "DataFrameSetitem"):
        target = ctx[op.target.key]
        # only deep copy when updating
        indexes = (
            (op.indexes,)
            if not isinstance(op.indexes, (tuple, list, set))
            else op.indexes
        )
        deep = bool(set(indexes) & set(target.columns))
        target = ctx[op.target.key].copy(deep=deep)
        value = ctx[op.value.key] if not np.isscalar(op.value) else op.value
        try:
            target[op.indexes] = value
        except KeyError:
            if _allow_set_missing_list:  # pragma: no cover
                raise
            else:
                existing = set(target.columns)
                new_columns = target.columns.append(
                    pd.Index([idx for idx in op.indexes if idx not in existing])
                )
                target = target.reindex(new_columns, axis=1)
                target[op.indexes] = value

        ctx[op.outputs[0].key] = target


def dataframe_setitem(df, col, value):
    op = DataFrameSetitem(target=df, indexes=col, value=value)
    return op(df, value)


def series_setitem(series, index, value):
    """
    Currently only supports series whose indexes contain `index`
    """
    op = DataFrameSetitem(target=series, indexes=index, value=value)
    return op(series, value)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import xgboost
except ImportError:  # pragma: no cover
    xgboost = None

import pytest

from ... import xgboost as xgb


@pytest.mark.skipif(xgboost is None, reason="XGBoost not installed")
def test_xgbclassifier_docstring():
    docstring = xgb.XGBClassifier.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from xgboost."
    )

    docstring = xgb.XGBClassifier.fit.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from xgboost.sklearn.XGBClassifier."
    )


@pytest.mark.skipif(xgboost is None, reason="XGBoost not installed")
def test_xgbregressor_docstring():
    docstring = xgb.XGBRegressor.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from xgboost."
    )

    docstring = xgb.XGBRegressor.fit.__doc__
    assert docstring is not None and docstring.endswith(
        "This docstring was copied from xgboost.sklearn.XGBRegressor."
    )



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import time
import uuid
from typing import Collection, Dict, List, Optional, Union

from ..._mars.deploy.utils import wait_services_ready
from ..._mars.lib.aio import new_isolation, stop_isolation
from ..._mars.services.cluster.api import WebClusterAPI
from ..._mars.session import new_session
from ..._mars.utils import calc_size_by_str
from ...compat._constants import SUPPORTED_PY_VERSIONS
from ...utils import get_local_py_version
from .config import (
    IngressConfig,
    NamespaceConfig,
    RoleBindingConfig,
    RoleConfig,
    ServiceConfig,
    XorbitsReplicationConfig,
    XorbitsSupervisorsConfig,
    XorbitsWorkersConfig,
)
from .external_storage.external_storage import JuicefsK8SStorage

try:
    from kubernetes.client import ApiClient
    from kubernetes.client.rest import ApiException as K8SApiException
except ImportError:  # pragma: no cover
    K8SApiException = None
    ApiClient = None

logger = logging.getLogger(__name__)


class KubernetesClusterClient:
    def __init__(self, cluster: "KubernetesCluster"):
        self._cluster = cluster
        self._endpoint = None
        self._session = None

    @property
    def endpoint(self) -> Optional[str]:
        return self._endpoint

    @property
    def namespace(self) -> Optional[str]:
        return self._cluster.namespace

    @property
    def session(self) -> Optional[str]:
        return self._session

    def start(self):
        try:
            self._endpoint = self._cluster.start()
            self._session = new_session(self._endpoint)
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except   # pragma: no cover
            self.stop()
            raise

    def stop(self, wait=False, timeout=0):
        self._cluster.stop(wait=wait, timeout=timeout)


class KubernetesCluster:
    _supervisor_config_cls = XorbitsSupervisorsConfig
    _worker_config_cls = XorbitsWorkersConfig
    _default_service_port: int = 7103
    _default_web_port: int = 7104

    def __init__(
        self,
        kube_api_client: ApiClient,
        worker_cpu: int,
        worker_mem: str,
        image: Optional[str] = None,
        namespace: Optional[str] = None,
        supervisor_num: int = 1,
        supervisor_cpu: int = 1,
        supervisor_mem: str = "4G",
        supervisor_mem_limit_ratio: Optional[float] = None,
        worker_num: int = 1,
        worker_spill_paths: Optional[str] = None,
        worker_cache_mem: Optional[str] = None,
        min_worker_num: Optional[int] = None,
        worker_min_cache_mem: Optional[str] = None,
        worker_mem_limit_ratio: Optional[float] = None,
        web_port: Optional[int] = None,
        service_name: Optional[str] = None,
        service_type: Optional[str] = None,
        pip: Optional[Union[str, List[str]]] = None,
        conda: Optional[Union[str, List[str]]] = None,
        timeout: Optional[int] = None,
        cluster_type: str = "auto",
        external_storage: Optional[str] = None,
        external_storage_config: Optional[dict] = None,
        **kwargs,
    ):
        from kubernetes import client as kube_client

        if worker_cpu is None or worker_mem is None:  # pragma: no cover
            raise TypeError("`worker_cpu` and `worker_mem` must be specified")
        if cluster_type not in ["auto", "eks", "kubernetes"]:  # pragma: no cover
            raise ValueError("`cluster_type` must be `auto`, `kubernetes` or `eks`")
        if pip is not None and not isinstance(pip, (str, list)):  # pragma: no cover
            raise TypeError("`pip` must be str or List[str] type.")
        if conda is not None and not isinstance(conda, (str, list)):  # pragma: no cover
            raise TypeError("`conda` must be str or List[str] type.")
        local_py_version = get_local_py_version()
        if (
            image is None and local_py_version not in SUPPORTED_PY_VERSIONS
        ):  # pragma: no cover
            raise RuntimeError(
                f"Xorbits does not support Kubernetes deployment under your python version {local_py_version}. "
                f"Please change the python version to one of the following versions: {SUPPORTED_PY_VERSIONS}."
            )

        self._api_client = kube_api_client
        self._core_api = kube_client.CoreV1Api(kube_api_client)
        self._networking_api = kube_client.NetworkingV1Api(self._api_client)

        self._namespace = namespace
        self._image = image
        self._timeout = timeout
        self._service_name = service_name or "xorbits-service"
        self._service_type = service_type or "NodePort"
        self._extra_volumes = kwargs.pop("extra_volumes", ())
        self._pre_stop_command = kwargs.pop("pre_stop_command", None)
        self._log_when_fail = kwargs.pop("log_when_fail", False)
        self._cluster_type = self._get_cluster_type(cluster_type)
        self._ingress_name = "xorbits-ingress"
        self._use_local_image = kwargs.pop("use_local_image", False)
        self._external_storage = external_storage
        self._external_storage_config = external_storage_config

        if self._external_storage and self._external_storage not in ["juicefs"]:
            raise ValueError(
                "Currently, only juicefs is supported as one of our storage backend."
            )
        if self._external_storage == "juicefs":
            if (
                not self._external_storage_config
                or "metadata_url" not in self._external_storage_config
            ):
                raise ValueError(
                    "For external storage JuiceFS, you must specify the metadata url for its metadata storage in external storage config, for example external_storage_config={'metadata_url': 'redis://172.17.0.5:6379/1',}."
                )

        extra_modules = kwargs.pop("extra_modules", None) or []
        extra_modules = (
            extra_modules.split(",")
            if isinstance(extra_modules, str)
            else extra_modules
        )
        extra_envs = kwargs.pop("extra_env", None) or dict()
        extra_labels = kwargs.pop("extra_labels", None) or dict()
        service_port = kwargs.pop("service_port", None) or self._default_service_port

        def _override_modules(updates: Union[str, Collection]):
            modules = set(extra_modules)
            updates = updates.split(",") if isinstance(updates, str) else updates
            modules.update(updates)
            return sorted(modules)

        def _override_dict(d, updates):
            updates = updates or dict()
            ret = d.copy()
            ret.update(updates)
            return ret

        _override_envs = functools.partial(_override_dict, extra_envs)
        _override_labels = functools.partial(_override_dict, extra_labels)

        self._supervisor_num = supervisor_num
        self._supervisor_cpu = supervisor_cpu
        self._supervisor_mem = calc_size_by_str(supervisor_mem, None)
        self._supervisor_mem_limit_ratio = supervisor_mem_limit_ratio
        self._supervisor_extra_modules = _override_modules(
            kwargs.pop("supervisor_extra_modules", [])
        )
        self._supervisor_extra_env = _override_envs(
            kwargs.pop("supervisor_extra_env", None)
        )
        self._supervisor_extra_labels = _override_labels(
            kwargs.pop("supervisor_extra_labels", None)
        )
        self._supervisor_service_port = (
            kwargs.pop("supervisor_service_port", None) or service_port
        )
        self._web_port = web_port or self._default_web_port
        self._external_web_endpoint = None

        self._worker_num = worker_num
        self._worker_cpu = worker_cpu
        self._worker_mem = calc_size_by_str(worker_mem, None)
        self._worker_mem_limit_ratio = worker_mem_limit_ratio
        self._worker_spill_paths = worker_spill_paths
        self._worker_cache_mem = worker_cache_mem
        self._worker_min_cache_men = worker_min_cache_mem
        self._min_worker_num = min_worker_num
        self._worker_extra_modules = _override_modules(
            kwargs.pop("worker_extra_modules", [])
        )
        self._worker_extra_env = _override_envs(kwargs.pop("worker_extra_env", None))
        self._worker_extra_labels = _override_labels(
            kwargs.pop("worker_extra_labels", None)
        )
        self._worker_service_port = (
            kwargs.pop("worker_service_port", None) or service_port
        )
        self._pip = pip
        self._conda = conda
        self._readiness_service_name = kwargs.pop(
            "readiness_service_name", "xorbits-readiness-service"
        )

    @property
    def namespace(self) -> Optional[str]:
        return self._namespace

    @staticmethod
    def _get_k8s_context(context: Dict) -> str:
        cluster = context.get("context", {}).get("cluster", "")
        user = context.get("context", {}).get("user", "")
        name = context.get("name", "")

        for s in [cluster, user, name]:
            if ("eks" in s) or ("aws" in s):
                return "eks"
        return "kubernetes"

    @staticmethod
    def _get_cluster_type(cluster_type: str) -> str:
        from kubernetes import config

        if cluster_type == "auto":
            try:
                curr_context = config.list_kube_config_contexts()[1]
                return KubernetesCluster._get_k8s_context(curr_context)
            except:  # pragma: no cover
                raise ValueError(
                    "`cluster_type` cannot be `auto`, please check your kubectl configuration or specify it as `eks` or `kubernetes`"
                )
        else:
            return cluster_type

    def _get_free_namespace(self) -> str:
        while True:
            namespace = "xorbits-ns-" + str(uuid.uuid4().hex)
            try:
                self._core_api.read_namespace(namespace)
            except K8SApiException as ex:
                if ex.status != 404:  # pragma: no cover
                    raise
                return namespace

    def _create_kube_service(self):
        if self._service_type != "NodePort":  # pragma: no cover
            raise NotImplementedError(
                f"Service type {self._service_type} not supported"
            )

        service_config = ServiceConfig(
            self._service_name,
            service_type="NodePort",
            port=self._web_port,
            selector={"xorbits/service-type": XorbitsSupervisorsConfig.rc_name},
        )
        self._core_api.create_namespaced_service(
            self._namespace, service_config.build()
        )

    def _create_readiness_service(self):
        """
        Start a simple ClusterIP service for the workers to detect whether the supervisor is successfully started.
        """
        readiness_service_config = ServiceConfig(
            name=self._readiness_service_name,
            service_type=None,
            selector={"xorbits/service-type": XorbitsSupervisorsConfig.rc_name},
            port=XorbitsReplicationConfig.default_readiness_port,
            target_port=XorbitsReplicationConfig.default_readiness_port,
            protocol="TCP",
        )
        self._core_api.create_namespaced_service(
            self._namespace, readiness_service_config.build()
        )

    def _get_ready_pod_count(self, label_selector: str) -> int:  # pragma: no cover
        query = self._core_api.list_namespaced_pod(
            namespace=self._namespace, label_selector=label_selector
        ).to_dict()
        cnt = 0
        for el in query["items"]:
            if el["status"]["phase"] in ("Error", "Failed"):
                logger.warning(
                    "Error in starting pod, message: %s", el["status"]["message"]
                )
                continue
            if "status" not in el or "conditions" not in el["status"]:
                cnt += 1
            elif any(
                cond["type"] == "Ready" and cond["status"] == "True"
                for cond in el["status"].get("conditions") or ()
            ):
                cnt += 1
        return cnt

    def _create_namespace(self):
        namespace = self._namespace = self._get_free_namespace()
        self._core_api.create_namespace(NamespaceConfig(namespace).build())

    def _create_roles_and_bindings(self):
        # create role and binding
        role_config = RoleConfig(
            "xorbits-pod-operator",
            self._namespace,
            api_groups="",
            resources="pods,endpoints,services",
            verbs="get,watch,list,patch",
        )
        role_config.create_namespaced(self._api_client, self._namespace)
        role_binding_config = RoleBindingConfig(
            "xorbits-pod-operator-binding",
            self._namespace,
            "xorbits-pod-operator",
            "default",
        )
        role_binding_config.create_namespaced(self._api_client, self._namespace)

    def _create_supervisors(self):
        supervisors_config = self._supervisor_config_cls(
            self._supervisor_num,
            image=self._image,
            cpu=self._supervisor_cpu,
            memory=self._supervisor_mem,
            memory_limit_ratio=self._supervisor_mem_limit_ratio,
            modules=self._supervisor_extra_modules,
            volumes=self._extra_volumes,
            service_name=self._service_name,
            service_port=self._supervisor_service_port,
            web_port=self._web_port,
            pre_stop_command=self._pre_stop_command,
            use_local_image=self._use_local_image,
            pip=self._pip,
            conda=self._conda,
            external_storage=self._external_storage,
            external_storage_config=self._external_storage_config,
        )
        supervisors_config.add_simple_envs(self._supervisor_extra_env)
        supervisors_config.add_labels(self._supervisor_extra_labels)
        supervisors_config.create_namespaced(self._api_client, self._namespace)

    def _create_workers(self):
        workers_config = self._worker_config_cls(
            self._worker_num,
            image=self._image,
            cpu=self._worker_cpu,
            memory=self._worker_mem,
            memory_limit_ratio=self._worker_mem_limit_ratio,
            spill_volumes=self._worker_spill_paths,
            modules=self._worker_extra_modules,
            volumes=self._extra_volumes,
            worker_cache_mem=self._worker_cache_mem,
            min_cache_mem=self._worker_min_cache_men,
            service_name=self._service_name,
            service_port=self._worker_service_port,
            pre_stop_command=self._pre_stop_command,
            supervisor_web_port=self._web_port,
            use_local_image=self._use_local_image,
            pip=self._pip,
            conda=self._conda,
            readiness_service_name=self._readiness_service_name,
            external_storage=self._external_storage,
            external_storage_config=self._external_storage_config,
        )
        workers_config.add_simple_envs(self._worker_extra_env)
        workers_config.add_labels(self._worker_extra_labels)
        workers_config.create_namespaced(self._api_client, self._namespace)

    def _create_services(self):
        self._create_supervisors()
        self._create_workers()

    def _wait_services_ready(self):
        min_worker_num = int(self._min_worker_num or self._worker_num)
        limits = [self._supervisor_num, min_worker_num]
        selectors = [
            "xorbits/service-type=" + XorbitsSupervisorsConfig.rc_name,
            "xorbits/service-type=" + XorbitsWorkersConfig.rc_name,
        ]
        start_time = time.time()
        logger.debug("Start waiting pods to be ready")
        wait_services_ready(
            selectors,
            limits,
            lambda sel: self._get_ready_pod_count(sel),
            timeout=self._timeout,
        )
        logger.info("All service pods ready.")
        if self._timeout is not None:  # pragma: no branch
            self._timeout -= time.time() - start_time

    def _create_ingress(self):
        self._networking_api.create_namespaced_ingress(
            self._namespace,
            IngressConfig(
                self._namespace,
                self._ingress_name,
                self._service_name,
                self._web_port,
                self._cluster_type,
            ).build(),
        )

    def _get_ingress_address(self):
        from kubernetes import watch

        w = watch.Watch()
        for event in w.stream(
            func=self._networking_api.list_namespaced_ingress, namespace=self._namespace
        ):
            if (
                event["object"].kind == "Ingress"
                and event["object"].status.load_balancer.ingress is not None
                and len(event["object"].status.load_balancer.ingress) > 0
            ):
                ingress = event["object"].status.load_balancer.ingress[0]
                ip = ingress.ip
                host = ingress.hostname if hasattr(ingress, "hostname") else None
                address = ip or host
                logger.debug(f"Ingress ip: {ip}, host: {host}, address: {address}")
                if address is not None:
                    return f"http://{address}:80"

    def _wait_web_ready(self):
        loop = new_isolation().loop

        async def get_supervisors():
            start_time = time.time()
            while True:
                try:
                    time.sleep(1)
                    # TODO use Xorbits API instead of mars.
                    cluster_api = WebClusterAPI(self._external_web_endpoint)
                    supervisors = await cluster_api.get_supervisors()

                    if len(supervisors) == self._supervisor_num:
                        break
                except:  # noqa: E722  # nosec  # pylint: disable=bare-except  # pragma: no cover
                    if (
                        self._timeout is not None
                        and time.time() - start_time > self._timeout
                    ):
                        logger.exception("Error when fetching supervisors")
                        raise TimeoutError(
                            "Wait for kubernetes cluster timed out"
                        ) from None

        asyncio.run_coroutine_threadsafe(get_supervisors(), loop).result()

    def _load_cluster_logs(self) -> Dict:  # pragma: no cover
        log_dict = dict()
        pod_items = self._core_api.list_namespaced_pod(self._namespace).to_dict()
        for item in pod_items["items"]:
            log_dict[item["metadata"]["name"]] = self._core_api.read_namespaced_pod_log(
                name=item["metadata"]["name"], namespace=self._namespace
            )
        return log_dict

    def _create_external_storage(self, storage_name: str):
        if storage_name == "juicefs":
            juicefs_k8s_storage = JuicefsK8SStorage(
                namespace=self.namespace,
                api_client=self._api_client,
                external_storage_config=self._external_storage_config,
            )
            juicefs_k8s_storage.build()

    def start(self):
        try:
            self._create_namespace()
            self._create_roles_and_bindings()

            if self._external_storage:
                self._create_external_storage(storage_name=self._external_storage)

            self._create_services()
            self._create_kube_service()
            self._create_readiness_service()

            self._wait_services_ready()

            self._create_ingress()

            self._external_web_endpoint = self._get_ingress_address()
            self._wait_web_ready()
            logger.warning(f"Xorbits endpoint {self._external_web_endpoint} is ready!")
            return self._external_web_endpoint
        except:  # noqa: E722   # pragma: no cover
            if self._log_when_fail:
                logger.error("Error when creating cluster")
                for name, log in self._load_cluster_logs().items():
                    logger.error("Error logs for %s:\n%s", name, log)
            self.stop()
            raise

    def _delete_pv(self):
        from kubernetes.client import CoreV1Api

        api = CoreV1Api(self._api_client)
        pv_name = f"juicefs-pv-{self._namespace}"
        try:
            # K8s issue: https://github.com/kubernetes/kubernetes/issues/69697
            api.patch_persistent_volume(
                name=pv_name, body={"metadata": {"finalizers": None}}
            )
            api.delete_persistent_volume(pv_name)
        except Exception as e:  # pragma: no cover
            logger.warning(f"Error happens during deleting pv: {e}")

    def stop(self, wait: bool = False, timeout: int = 0):
        # stop isolation
        stop_isolation()

        from kubernetes.client import CoreV1Api

        api = CoreV1Api(self._api_client)
        api.delete_namespace(self._namespace)
        if self._external_storage == "juicefs":
            self._delete_pv()
        if wait:
            start_time = time.time()
            while True:
                try:
                    api.read_namespace(self._namespace)
                except K8SApiException as ex:  # pragma: no cover
                    if ex.status != 404:
                        raise
                    break
                else:
                    time.sleep(1)
                    if (
                        timeout and time.time() - start_time > timeout
                    ):  # pragma: no cover
                        raise TimeoutError


def new_cluster(
    kube_api_client: ApiClient,
    worker_cpu: int,
    worker_mem: str,
    image: Optional[str] = None,
    supervisor_num: int = 1,
    supervisor_cpu: int = 1,
    supervisor_mem: str = "4G",
    worker_num: int = 1,
    worker_spill_paths: Optional[List[str]] = None,
    worker_cache_mem: Optional[str] = None,
    min_worker_num: Optional[int] = None,
    pip: Optional[Union[str, List[str]]] = None,
    conda: Optional[Union[str, List[str]]] = None,
    timeout: Optional[int] = None,
    cluster_type: str = "auto",
    external_storage: Optional[str] = None,
    **kwargs,
) -> "KubernetesClusterClient":
    """
    The entrance of deploying xorbits cluster.

    Parameters
    ----------
    kube_api_client :
        Kubernetes API client, required, can be created with ``new_client_from_config``
    worker_cpu :
        Number of CPUs for every worker, required
    worker_mem :
        Memory size for every worker, required
    image :
        Docker image to use, ``xprobe/xorbits:<xorbits version>`` by default
    supervisor_num :
        Number of supervisors in the cluster, 1 by default
    supervisor_cpu :
        Number of CPUs for every supervisor, 1 by default
    supervisor_mem :
        Memory size for every supervisor, 4G by default
    worker_num :
        Number of workers in the cluster, 1 by default
    worker_spill_paths :
        Spill paths for worker pods on host
    worker_cache_mem :
        Size or ratio of cache memory for every worker
    min_worker_num :
        Minimal ready workers, equal to ``worker_num`` by default
    pip :
        Either a list of pip requirements specifiers,
        or a string containing the path to a pip `requirements.txt <https://pip.pypa.io/en/stable/user_guide/#requirements-files>`_ file.
        None by default.
        Both supervisor and worker will install the specified pip packages.
        Examples:

            * ``pip=["requests==1.0.0", "aiohttp"]``
            * ``pip="/path/to/requirements.txt"``
    conda :
        Either a string containing the path to a `conda environment.yml <https://conda.io/projects/conda/en/latest/user-guide/tasks/manage-environments.html#create-env-file-manually>`_ file,
        or a list of conda packages used for `conda install <https://docs.conda.io/projects/conda/en/latest/commands/install.html>`_ command.
        None by default.
        Both supervisor and worker will install the specified conda packages.
        When this parameter is list type, install the conda packages from the `default channel <https://repo.anaconda.com/pkgs/>`_.
        When this parameter is string type, the ``name`` attribute in the environment.yml file will not take effect,
        since all package changes will occur in the ``base`` conda environment where Xorbits exists.
        Examples:

            * ``conda=["tensorflow", "tensorboard"]``
            * ``conda="/path/to/environment.yml"``
    timeout :
        Timeout in seconds when creating clusters, never timeout by default
    cluster_type :
        K8s cluster type, ``auto``, ``kubernetes`` or ``eks`` supported, ``auto`` by default.
        ``auto`` means that it will automatically detect whether the kubectl context is ``eks``.
        You can also manually specify ``kubernetes`` or ``eks`` in some special cases.
    external_storage:
        You can specify an external storage. Currently, only juicefs is supported as one of our storage backend.
    kwargs :
        Extra kwargs

    Returns
    -------
    KubernetesClusterClient
        a KubernetesClusterClient instance
    """
    cluster_cls = kwargs.pop("cluster_cls", KubernetesCluster)

    cluster = cluster_cls(
        kube_api_client,
        worker_cpu,
        worker_mem,
        image=image,
        supervisor_num=supervisor_num,
        supervisor_cpu=supervisor_cpu,
        supervisor_mem=supervisor_mem,
        worker_num=worker_num,
        worker_spill_paths=worker_spill_paths,
        worker_cache_mem=worker_cache_mem,
        min_worker_num=min_worker_num,
        pip=pip,
        conda=conda,
        timeout=timeout,
        cluster_type=cluster_type,
        external_storage=external_storage,
        **kwargs,
    )
    client = KubernetesClusterClient(cluster)
    client.start()
    return client



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    import jax
    from jax import config

    config.update("jax_enable_x64", True)
    JAX_INSTALLED = True
except ImportError:
    JAX_INSTALLED = False

from .. import arithmetic, reduction
from ..operands import TensorFuse
from .core import TensorFuseChunkMixin


class TensorJAXFuseChunk(TensorFuse, TensorFuseChunkMixin):
    _op_type_ = None  # no opcode, cannot be serialized

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        inputs = [ctx[c.key] for c in op.inputs]
        input_dict = {c.key: i for c, i in zip(op.inputs, inputs)}
        res = np.array(_evaluate(chunk, input_dict))
        res = _maybe_keepdims(chunk, res)
        if chunk.ndim == 0 and res.ndim == 1 and res.size == 0:
            res = res.dtype.type(0)
        ctx[chunk.key] = res


# recursively compute the inputs
def _evaluate(chunk, input_dict):
    op_type = type(chunk.op)

    if chunk.key in input_dict:
        return input_dict[chunk.key]
    elif op_type is TensorJAXFuseChunk:
        return _evaluate(chunk.composed[-1], input_dict)
    elif op_type in ARITHMETIC_SUPPORT:
        if hasattr(chunk.op, "inputs"):
            if hasattr(chunk.op, "input"):
                return _get_jax_function(chunk.op)(
                    _evaluate(chunk.op.input, input_dict)
                )
            elif np.isscalar(chunk.op.lhs):
                rhs = _evaluate(chunk.op.rhs, input_dict)
                return _get_jax_function(chunk.op)(chunk.op.lhs, rhs)
            elif np.isscalar(chunk.op.rhs):
                lhs = _evaluate(chunk.op.lhs, input_dict)
                return _get_jax_function(chunk.op)(lhs, chunk.op.rhs)
            else:
                lhs = _evaluate(chunk.op.lhs, input_dict)
                rhs = _evaluate(chunk.op.rhs, input_dict)
                return _get_jax_function(chunk.op)(lhs, rhs)
        return _get_jax_function(chunk.op)
    elif op_type in TREE_SUPPORT:
        func = _get_jax_function(chunk.op)
        inputs = [_evaluate(input, input_dict) for input in chunk.inputs]

        # passing func and inputs as parameters seems to cause error with jax.jit
        def _fusion():
            output = func(inputs[0], inputs[1])
            for input in inputs[2:]:
                output = func(output, input)
            return output

        return jax.jit(_fusion)()
    elif op_type in REDUCTION_SUPPORT:
        ax = chunk.op.axis
        data = _evaluate(chunk.inputs[0], input_dict)
        if len(ax) == data.ndim:
            return _get_jax_function(chunk.op)(data)
        else:
            return _get_jax_function(chunk.op)(data, axis=ax)
    else:
        raise TypeError(f"unsupported operator in jax: {op_type.__name__}")


def _get_jax_function(operand):
    func_name = getattr(operand, "_func_name")
    return getattr(jax.numpy, func_name)


def _maybe_keepdims(chunk, res):
    out_chunk = chunk.composed[-1] if type(chunk.op) == TensorJAXFuseChunk else chunk
    if type(out_chunk.op) in REDUCTION_SUPPORT and out_chunk.op.keepdims:
        res = np.reshape(res, out_chunk.shape)
    return res


ARITHMETIC_SUPPORT = {
    arithmetic.TensorNegative,
    arithmetic.TensorAbs,
    arithmetic.TensorConj,
    arithmetic.TensorExp,
    arithmetic.TensorLog,
    arithmetic.TensorLog10,
    arithmetic.TensorExpm1,
    arithmetic.TensorLog1p,
    arithmetic.TensorSqrt,
    arithmetic.TensorSin,
    arithmetic.TensorCos,
    arithmetic.TensorTan,
    arithmetic.TensorArcsin,
    arithmetic.TensorArccos,
    arithmetic.TensorArctan,
    arithmetic.TensorSinh,
    arithmetic.TensorCosh,
    arithmetic.TensorTanh,
    arithmetic.TensorArcsinh,
    arithmetic.TensorArccosh,
    arithmetic.TensorArctanh,
    arithmetic.TensorFloor,
    arithmetic.TensorCeil,
    arithmetic.TensorNot,
    arithmetic.TensorAdd,
    arithmetic.TensorSubtract,
    arithmetic.TensorMultiply,
    arithmetic.TensorDivide,
    arithmetic.TensorMod,
    arithmetic.TensorPower,
    arithmetic.TensorLshift,
    arithmetic.TensorRshift,
    arithmetic.TensorEqual,
    arithmetic.TensorNotEqual,
    arithmetic.TensorLessThan,
    arithmetic.TensorLessEqual,
    arithmetic.TensorGreaterThan,
    arithmetic.TensorGreaterEqual,
    arithmetic.TensorAnd,
    arithmetic.TensorOr,
}

TREE_SUPPORT = {
    arithmetic.TensorTreeAdd,
    arithmetic.TensorTreeMultiply,
}

REDUCTION_SUPPORT = {
    reduction.TensorSum,
    reduction.TensorProd,
    reduction.TensorMax,
    reduction.TensorMin,
}



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

MARS_TEMP_DIR = str(Path.home() / ".xorbits")

# logging.
MARS_LOG_DIR_KEY = "MARS_LOG_DIR"
DEFAULT_MARS_LOG_DIR = os.path.join(MARS_TEMP_DIR, "logs")
DEFAULT_MARS_LOG_FILE_NAME = "xorbits.log"
DEFAULT_MARS_LOG_MAX_BYTES = 100 * 1024 * 1024
DEFAULT_MARS_LOG_BACKUP_COUNT = 30

# profiling.
MARS_PROFILING_RESULTS_DIR = os.path.join(MARS_TEMP_DIR, "profiling")



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from types import ModuleType
from typing import Callable, Dict, Set, Type

import pandas

from ...core.adapter import (
    MARS_DATAFRAME_GROUPBY_TYPE,
    MARS_DATAFRAME_OR_SERIES_TYPE,
    MARS_DATAFRAME_TYPE,
    MARS_SERIES_GROUPBY_TYPE,
    MARS_SERIES_TYPE,
    MarsDataFrameToCSV,
    MarsDataFrameToParquet,
    MarsDataFrameToSQLTable,
    MarsDataFrameToVineyardChunk,
    MarsEntity,
    collect_cls_members,
    from_mars,
    mars_dataframe,
    own_data,
    register_from_mars_execution_condition,
    register_to_mars_execution_condition,
    to_mars,
    wrap_mars_callable,
)
from ...core.data import DataType
from ...core.utils.docstring import attach_cls_member_docstring


# functions and class constructors defined by mars dataframe
def _collect_module_callables() -> Dict[str, Callable]:
    mars_dataframe_callables = dict()

    # install class constructors.
    mars_dataframe_callables[mars_dataframe.DataFrame.__name__] = wrap_mars_callable(
        mars_dataframe.DataFrame,
        attach_docstring=True,
        is_cls_member=False,
        docstring_src_module=pandas,
        docstring_src=pandas.DataFrame,
    )
    mars_dataframe_callables[mars_dataframe.Series.__name__] = wrap_mars_callable(
        mars_dataframe.Series,
        attach_docstring=True,
        is_cls_member=False,
        docstring_src_module=pandas,
        docstring_src=pandas.Series,
    )
    mars_dataframe_callables[mars_dataframe.Index.__name__] = wrap_mars_callable(
        mars_dataframe.Index,
        attach_docstring=True,
        is_cls_member=False,
        docstring_src_module=pandas,
        docstring_src=pandas.Index,
    )
    # install module functions
    for name, func in inspect.getmembers(mars_dataframe, inspect.isfunction):
        mars_dataframe_callables[name] = wrap_mars_callable(
            func,
            attach_docstring=True,
            is_cls_member=False,
            docstring_src_module=pandas,
            docstring_src=getattr(pandas, name, None),
        )

    return mars_dataframe_callables


MARS_DATAFRAME_CALLABLES: Dict[str, Callable] = _collect_module_callables()


def _collect_dataframe_magic_methods() -> Set[str]:
    magic_methods = set()

    magic_methods_to_skip: Set[str] = {
        "__init__",
        "__dir__",
        "__repr__",
        "__str__",
        "__setattr__",
        "__getattr__",
        "__len__",
        "__array__",
    }
    all_cls = (
        MARS_DATAFRAME_TYPE
        + MARS_SERIES_TYPE
        + MARS_DATAFRAME_GROUPBY_TYPE
        + MARS_SERIES_GROUPBY_TYPE
    )
    for mars_cls in all_cls:
        for name, _ in inspect.getmembers(mars_cls, inspect.isfunction):
            if (
                name.startswith("__")
                and name.endswith("__")
                and name not in magic_methods_to_skip
            ):
                magic_methods.add(name)
    return magic_methods


MARS_DATAFRAME_MAGIC_METHODS: Set[str] = _collect_dataframe_magic_methods()


def _register_to_mars_execution_conditions() -> None:
    def _on_dtypes_being_none(mars_entity: "MarsEntity") -> bool:
        if hasattr(mars_entity, "dtypes") and mars_entity.dtypes is None:
            return True
        return False

    register_to_mars_execution_condition(
        mars_dataframe.DataFrame.__name__, _on_dtypes_being_none
    )


def _register_from_mars_execution_conditions() -> None:
    def _on_dataframe_export_functions_being_called(mars_entity: "MarsEntity") -> bool:
        return isinstance(
            mars_entity.op,
            (
                MarsDataFrameToParquet,
                MarsDataFrameToCSV,
                MarsDataFrameToSQLTable,
                MarsDataFrameToVineyardChunk,
            ),
        )

    def _on_series_export_functions_being_called(mars_entity: "MarsEntity") -> bool:
        return isinstance(mars_entity.op, (MarsDataFrameToCSV, MarsDataFrameToSQLTable))

    register_from_mars_execution_condition(
        mars_dataframe.DataFrame.__name__, _on_dataframe_export_functions_being_called
    )
    register_from_mars_execution_condition(
        mars_dataframe.Series.__name__, _on_series_export_functions_being_called
    )


def install_members(
    cls: Type, mars_cls: Type, docstring_src_module: ModuleType, docstring_src_cls: Type
):
    members = collect_cls_members(
        mars_cls,
        docstring_src_module=docstring_src_module,
        docstring_src_cls=docstring_src_cls,
    )
    for name in members:
        setattr(cls, name, members[name])


def wrap_user_defined_functions(
    c: Callable, member_name: str, data_type: DataType
) -> Callable:
    """
    A function wrapper for user defined functions.
    """

    @functools.wraps(c)
    def wrapped(*args, **kwargs):
        new_args = to_mars(args)
        new_kwargs = to_mars(kwargs)
        try:
            return from_mars(c(*new_args, **new_kwargs))
        except (TypeError, ValueError):
            # infer failed, add skip_infer=True manually
            new_kwargs["skip_infer"] = True
            ret = c(*new_args, **new_kwargs)
            if isinstance(ret, MARS_DATAFRAME_OR_SERIES_TYPE):
                ret = ret.ensure_data()
            return from_mars(ret)

    return attach_cls_member_docstring(wrapped, member_name, data_type=data_type)


def wrap_iteration_functions(
    member_method: Callable,
    member_name: str,
    data_type: DataType,
    attach_docstring: bool,
):
    """
    Wrapper for iteration functions.
    """

    @functools.wraps(member_method)
    def wrapped(self: MarsEntity, *args, **kwargs):
        if own_data(self):
            # if own data, return iteration on data directly
            return getattr(self.op.data, member_name)(*to_mars(args), **to_mars(kwargs))
        else:
            return from_mars(member_method(self, *to_mars(args), **to_mars(kwargs)))

    if attach_docstring:
        return attach_cls_member_docstring(wrapped, member_name, data_type)
    else:  # pragma: no cover
        wrapped.__doc__ = ""
        return wrapped



import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...serialization.serializables import Int64Field, ListField, StringField
from ..core import IndexValue
from ..merge.concat import DataFrameConcat
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_concatenated_rows_frame, parse_index


class DataFrameNLargest(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.NLARGEST

    largestOrSmallest = StringField("largestOrSmallest", default=None)
    n = Int64Field("n", default=0)
    columns = ListField("columns", default=None)
    keep = StringField("keep", default="first")

    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)

    @classmethod
    def _tile_dataframe(cls, op: "DataFrameNLargest"):
        df = op.inputs[0]
        df = build_concatenated_rows_frame(df)
        input_chunks = df.chunks

        if op.n >= input_chunks[0].shape[0]:
            out_chunks = input_chunks
        else:
            out_chunks = []
            for chunk in input_chunks:
                chunk_op = op.copy().reset_key()
                chunk_params = chunk.params
                if op.keep == "all":
                    chunk_params["shape"] = (np.nan, input_chunks[0].shape[1])
                else:
                    n = max(chunk.shape[0], op.n)
                    chunk_params["shape"] = (n, input_chunks[0].shape[1])
                chunk_params["index_value"] = parse_index(
                    df.chunks[0].index_value.to_pandas()[:0]
                )
                out_chunks.append(chunk_op.new_chunk([chunk], kws=[chunk_params]))
        op_concat = DataFrameConcat(
            axis=0,
            output_types=[OutputType.dataframe],
        )
        if op.keep == "all":
            shape_concat = (np.nan, input_chunks[0].shape[1])
        else:
            shape_concat = (
                sum(c.shape[0] for c in out_chunks),
                input_chunks[0].shape[1],
            )
        chunk_concat = op_concat.new_chunk(
            out_chunks,
            shape=shape_concat,
            index_value=parse_index(df.chunks[0].index_value.to_pandas()[:0]),
        )
        final_op = op.copy().reset_key()
        chunk_params = input_chunks[0].params
        chunk_params["index_value"] = parse_index(
            df.chunks[0].index_value.to_pandas()[:0]
        )
        if op.keep == "all":
            chunk_params["shape"] = (np.nan, input_chunks[0].shape[1])
        else:
            chunk_params["shape"] = (op.n, input_chunks[0].shape[1])
        c = final_op.new_chunk([chunk_concat], kws=[chunk_params])
        new_op = op.copy()
        kws = op.outputs[0].params.copy()
        if op.keep == "all":
            kws["nsplits"] = ((np.nan,), (op.outputs[0].shape[1],))
        else:
            kws["nsplits"] = ((op.n,), (op.outputs[0].shape[1],))
        kws["chunks"] = [c]
        return new_op.new_dataframes(op.inputs, **kws)

    @classmethod
    def _tile_series(cls, op: "DataFrameNLargest"):
        inp = op.inputs[0]
        if op.n >= inp.chunks[0].shape[0]:
            out_chunks = inp.chunks
        else:
            out_chunks = []
            for chunk in inp.chunks:
                chunk_op = op.copy().reset_key()
                chunk_params = chunk.params
                if op.keep == "all":
                    chunk_params["shape"] = (np.nan,)
                else:
                    n = max(chunk.shape[0], op.n)
                    chunk_params["shape"] = (n,)
                chunk_params["index_value"] = parse_index(
                    chunk.index_value.to_pandas()[:0]
                )
                out_chunks.append(chunk_op.new_chunk([chunk], kws=[chunk_params]))

        op_concat = DataFrameConcat(
            axis=0,
            output_types=[OutputType.dataframe],
        )
        if op.keep == "all":
            shape_concat = (np.nan,)
        else:
            shape_concat = (sum(c.shape[0] for c in out_chunks),)
        chunk_concat = op_concat.new_chunk(
            out_chunks,
            shape=shape_concat,
            index_value=parse_index(inp.chunks[0].index_value.to_pandas()[:0]),
        )
        final_op = op.copy().reset_key()
        chunk_params = inp.chunks[0].params
        chunk_params["index_value"] = parse_index(
            inp.chunks[0].index_value.to_pandas()[:0]
        )
        if op.keep == "all":
            chunk_params["shape"] = (np.nan,)
        else:
            chunk_params["shape"] = (op.n,)
        c = final_op.new_chunk([chunk_concat], kws=[chunk_params])
        new_op = op.copy()
        params = op.outputs[0].params
        params["chunks"] = [c]
        if op.keep == "all":
            params["nsplits"] = ((np.nan,),)
        else:
            params["nsplits"] = ((op.n,),)
        return new_op.new_seriess(op.inputs, kws=[params])

    @classmethod
    def tile(cls, op: "DataFrameNLargest"):
        inp = op.inputs[0]
        if inp.ndim == 2:
            return cls._tile_dataframe(op)
        else:
            return cls._tile_series(op)

    @classmethod
    def execute(cls, ctx, op: "DataFrameNLargest"):
        in_data = ctx[op.inputs[0].key]
        if in_data.ndim == 2:
            if op.largestOrSmallest == "largest":
                result = in_data.nlargest(n=op.n, columns=op.columns, keep=op.keep)
            elif op.largestOrSmallest == "smallest":
                result = in_data.nsmallest(n=op.n, columns=op.columns, keep=op.keep)
        else:
            if op.largestOrSmallest == "largest":
                result = in_data.nlargest(n=op.n, keep=op.keep)
            elif op.largestOrSmallest == "smallest":
                result = in_data.nsmallest(n=op.n, keep=op.keep)
        ctx[op.outputs[0].key] = result

    def __call__(self, a):
        if self.n > a.shape[0]:
            self.n = a.shape[0]
        if a.ndim == 2:
            if self.keep == "all":
                return self.new_dataframe(
                    [a],
                    shape=(np.nan, a.shape[1]),
                    dtypes=a.dtypes,
                    columns_value=a.columns_value,
                )
            else:
                return self.new_dataframe(
                    [a],
                    shape=(self.n, a.shape[1]),
                    dtypes=a.dtypes,
                    columns_value=a.columns_value,
                )
        else:
            if isinstance(a.index_value.value, IndexValue.RangeIndex):
                index_value = parse_index(pd.Index([], dtype=np.int64))
            else:
                index_value = a.index_value
            if self.keep == "all":
                return self.new_series(
                    [a],
                    shape=(np.nan,),
                    dtype=a.dtype,
                    index_value=index_value,
                    name=a.name,
                )
            else:
                return self.new_series(
                    [a],
                    shape=(self.n,),
                    dtype=a.dtype,
                    index_value=index_value,
                    name=a.name,
                )


def dataframe_nlargest(df, n, columns, keep="first"):
    """
    Return the first n rows ordered by columns in descending order.

    Parameters
    ----------
    df : Mars DataFrame
         Input dataframe.
    n :  int
         Number of rows to return.
    columns :  label or list of labels
         Column label(s) to order by.
    keep{‘first’, ‘last’, ‘all’}, default ‘first’
        Where there are duplicate values:
            first : prioritize the first occurrence(s)
            last : prioritize the last occurrence(s)
            all : do not drop any duplicates, even it means selecting more than n items.

    Returns
    -------
    sorted_obj : Mars DataFrame
        The first n rows ordered by the given columns in descending order.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({
    ...     'col1': ['A', 'A', 'B', 'E', 'D', 'C'],
    ...     'col2': [2, 1, 9, 8, 7, 4],
    ...     'col3': [0, 1, 9, 4, 2, 3],
    ... })
    >>> df.execute()
        col1 col2 col3
    0   A    2    0
    1   A    1    1
    2   B    9    9
    3   E    8    2
    4   D    8    4
    5   C    4    3

    Choose the first 3 rows ordered by col2

    >>> df.nlargest(3,"col2").execute()
        col1 col2 col3
    2    B     9     9
    3    E     8     2
    4    D     8     4

    Choose the first 3 rows ordered by multiple columns

    >>> df.nlargest(3,['col2', 'col3']).execute()
        col1 col2 col3
    2    B     9     9
    4    E     8     4
    3    D     8     2
    """
    if keep not in ["last", "first", "all"]:
        raise ValueError(f'''keep must be either "first", "last" or "all"''')
    if isinstance(columns, str):
        columns = [columns]
    elif isinstance(columns, list):
        columns = columns
    else:
        raise KeyError(columns)

    op = DataFrameNLargest(
        largestOrSmallest="largest",
        n=n,
        columns=columns,
        keep=keep,
        output_types=[OutputType.dataframe],
        gpu=df.op.is_gpu(),
    )
    nlargest_df = op(df)
    return nlargest_df


def series_nlargest(series, n, keep="first"):
    """
    Return the largest n elements.

    Parameters
    ----------
    df : Mars Series
         Input Series.
    n :  int
         Number of rows to return.
    keep{‘first’, ‘last’, ‘all’}, default ‘first’
        Where there are duplicate values:
            first : prioritize the first occurrence(s)
            last : prioritize the last occurrence(s)
            all : do not drop any duplicates, even it means selecting more than n items.

    Returns
    -------
    sorted_obj : Mars Series
        The n largest values in the Series, sorted in decreasing order.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> raw = pd.Series([8, 1, 3, 10, 5])
    >>> df.execute()
    0     8.0
    1     1.0
    2     3.0
    3     10.0
    4     5.0
    dtype: float64

    Choose the largest 3 rows

    >>> df.nlargest(3).execute()
    3     10.0
    0     8.0
    4     5.0
    dtype: float64
    """
    if keep not in ["last", "first", "all"]:
        raise ValueError(f'''keep must be either "first", "last" or "all"''')
    op = DataFrameNLargest(
        largestOrSmallest="largest",
        n=n,
        keep=keep,
        output_types=[OutputType.series],
        gpu=series.op.is_gpu(),
    )
    nlargest_df = op(series)
    return nlargest_df



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import inspect
from typing import Any, Callable, Dict, Optional

from ...core.utils.fallback import unimplemented_func


def __dir__():
    from ..mars_adapters import MARS_TENSOR_FFT_CALLABLES
    from ..numpy_adapters.core import NUMPY_FFT_MEMBERS

    global NUMPY_FFT_METHODS
    import numpy

    return list(MARS_TENSOR_FFT_CALLABLES.keys()) + list(NUMPY_FFT_MEMBERS.keys())


def __getattr__(name: str):
    from ..mars_adapters import MARS_TENSOR_FFT_CALLABLES
    from ..numpy_adapters.core import NUMPY_FFT_MEMBERS

    if name in MARS_TENSOR_FFT_CALLABLES:
        return MARS_TENSOR_FFT_CALLABLES[name]
    else:  # pragma: no cover
        import numpy

        if not hasattr(numpy.fft, name):
            raise AttributeError(name)
        elif name in NUMPY_FFT_MEMBERS:
            return NUMPY_FFT_MEMBERS[name]
        else:
            if inspect.ismethod(getattr(numpy.fft, name)):
                return unimplemented_func
            else:
                raise AttributeError(name)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap
from functools import partial


def _get_prefix_suffix_docs(is_prefix: bool):
    if is_prefix:
        action, pos = "prefix", "before"
        r_action = "suffix"
    else:
        action, pos = "suffix", "after"
        r_action = "prefix"

    def mk_col(ch: str, s: str):
        return f"{ch}_{s}" if is_prefix else f"{s}_{ch}"

    doc = f"""
    {action.capitalize()} labels with string `{action}`.

    For Series, the row labels are {action}ed.
    For DataFrame, the column labels are {action}ed.

    Parameters
    ----------
    {action} : str
        The string to add {pos} each label.

    Returns
    -------
    Series or DataFrame
        New Series or DataFrame with updated labels.

    See Also
    --------
    Series.add_{r_action}: Suffix row labels with string `{r_action}`.
    DataFrame.add_{r_action}: Suffix column labels with string `{r_action}`.

    Examples
    --------
    >>> import mars.dataframe as md
    >>> s = md.Series([1, 2, 3, 4])
    >>> s.execute()
    0    1
    1    2
    2    3
    3    4
    dtype: int64

    >>> s.add_prefix({mk_col('item', '')!r}).execute()
    {mk_col('item', '0')}    1
    {mk_col('item', '1')}    2
    {mk_col('item', '2')}    3
    {mk_col('item', '3')}    4
    dtype: int64

    >>> df = md.DataFrame({{'A': [1, 2, 3, 4], 'B': [3, 4, 5, 6]}})
    >>> df.execute()
       A  B
    0  1  3
    1  2  4
    2  3  5
    3  4  6

    >>> df.add_prefix({mk_col('col', '')!r}).execute()
         {mk_col('col', 'A')}  {mk_col('col', 'B')}
    0        1      3
    1        2      4
    2        3      5
    3        4      6
    """
    return textwrap.dedent(doc).strip()


def df_add_prefix(df, prefix):
    f = partial("{prefix}{}".format, prefix=prefix)
    return df.rename(columns=f)


def series_add_prefix(series, prefix):
    f = partial("{prefix}{}".format, prefix=prefix)
    return series.rename(index=f)


def df_add_suffix(df, suffix):
    f = partial("{}{suffix}".format, suffix=suffix)
    return df.rename(columns=f)


def series_add_suffix(series, suffix):
    f = partial("{}{suffix}".format, suffix=suffix)
    return series.rename(index=f)


df_add_prefix.__doc__ = _get_prefix_suffix_docs(True)
series_add_prefix.__doc__ = df_add_prefix.__doc__
df_add_suffix.__doc__ = _get_prefix_suffix_docs(False)
series_add_suffix.__doc__ = df_add_suffix.__doc__



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import inspect
import warnings
from functools import lru_cache
from typing import Any, Callable, Dict, Type

import pandas as pd

from ..._mars.core import Entity as MarsEntity
from ...core.adapter import (
    ClsMethodWrapper,
    MarsOutputType,
    get_cls_members,
    wrap_mars_callable,
)
from ...core.data import DataType
from ...core.utils.fallback import wrap_fallback_module_method

_NO_ANNOTATION_FUNCS: Dict[Callable, MarsOutputType] = {
    pd.read_pickle: MarsOutputType.object,
    pd.crosstab: MarsOutputType.dataframe,
    pd.infer_freq: MarsOutputType.object,
    pd.interval_range: MarsOutputType.index,
    pd.json_normalize: MarsOutputType.dataframe,
    pd.lreshape: MarsOutputType.dataframe,
    pd.merge_asof: MarsOutputType.dataframe,
    pd.merge_ordered: MarsOutputType.dataframe,
    pd.period_range: MarsOutputType.index,
    pd.pivot: MarsOutputType.dataframe,
    pd.pivot_table: MarsOutputType.dataframe,
    pd.read_excel: MarsOutputType.dataframe,
    pd.read_fwf: MarsOutputType.dataframe,
    pd.read_gbq: MarsOutputType.dataframe,
    pd.read_hdf: MarsOutputType.object,
    pd.read_html: MarsOutputType.object,
    pd.read_json: MarsOutputType.df_or_series,
    pd.read_orc: MarsOutputType.dataframe,
    pd.read_sas: MarsOutputType.dataframe,
    pd.read_spss: MarsOutputType.dataframe,
    pd.read_stata: MarsOutputType.dataframe,
    pd.read_table: MarsOutputType.dataframe,
    pd.wide_to_long: MarsOutputType.dataframe,
}

if pd.__version__ >= "1.3.0":  # pragma: no branch
    _NO_ANNOTATION_FUNCS[pd.read_xml] = MarsOutputType.dataframe


class PandasClsMethodWrapper(ClsMethodWrapper):
    def _generate_fallback_data(self, mars_entity: MarsEntity):
        return mars_entity.to_pandas()

    def _generate_warning_msg(self, entity: MarsEntity, func_name: str):
        return f"{type(entity).__name__}.{func_name} will fallback to Pandas"

    def _get_output_type(self, func: Callable) -> MarsOutputType:
        return_annotation = inspect.signature(func).return_annotation
        # since pandas v2.1.0, the return_annotation of `pd.read_pickle` becomes `DataFrame | Series`,
        # see https://github.com/pandas-dev/pandas/blob/v2.1.0/pandas/io/pickle.py#L116-L214 .
        # However, the output_type should be `object` according to its document
        # (https://pandas.pydata.org/docs/reference/api/pandas.read_pickle.html).
        # So here's an extra check.
        if return_annotation is inspect.Signature.empty or func == pd.read_pickle:
            # mostly for python3.7 whose return_annotation is always empty
            return _NO_ANNOTATION_FUNCS.get(func, MarsOutputType.object)
        all_types = [t.strip() for t in return_annotation.split("|")]
        has_df = "DataFrame" in all_types
        has_series = "Series" in all_types
        if has_df and has_series:
            # output could be df or series, use OutputType.df_or_series
            output_type = MarsOutputType.df_or_series
        elif has_df:
            # output is df, use OutputType.dataframe
            output_type = MarsOutputType.dataframe
        elif has_series:
            # output is series, use OutputType.series
            output_type = MarsOutputType.series
        else:
            # otherwise, use object
            output_type = MarsOutputType.object
        return output_type

    def _get_docstring_src_module(self):
        return pd


class PandasFetchDataMethodWrapper(PandasClsMethodWrapper):
    def get_wrapped(self):
        wrapped = super().get_wrapped()

        @functools.wraps(getattr(self.library_cls, self.func_name))
        def _wrapped(entity: MarsEntity, *args, **kwargs):
            def fetch_wrapped(func):
                ret = func(entity, *args, **kwargs)
                return ret.fetch()

            return fetch_wrapped(wrapped)

        return _wrapped


def _collect_pandas_cls_members(pd_cls: Type, data_type: DataType):
    members = get_cls_members(data_type)
    for name, pd_cls_member in inspect.getmembers(pd_cls, inspect.isfunction):
        if name == "tolist" and pd_cls == pd.Series:
            pandas_series_tolist_method_wrapper = PandasFetchDataMethodWrapper(
                library_cls=pd_cls, func_name=name, fallback_warning=True
            )
            members[name] = pandas_series_tolist_method_wrapper.get_wrapped()
        elif name not in members and not name.startswith("_"):
            pandas_cls_method_wrapper = PandasClsMethodWrapper(
                library_cls=pd_cls, func_name=name, fallback_warning=True
            )
            members[name] = pandas_cls_method_wrapper.get_wrapped()
    # make to_numpy an alias of to_tensor
    members["to_numpy"] = members["to_tensor"]


def _collect_pandas_dataframe_members():
    _collect_pandas_cls_members(pd.DataFrame, DataType.dataframe)


def _collect_pandas_series_members():
    _collect_pandas_cls_members(pd.Series, DataType.series)


def _collect_pandas_index_members():
    _collect_pandas_cls_members(pd.Index, DataType.index)


@lru_cache(maxsize=1)
def collect_pandas_module_members() -> Dict[str, Any]:
    from ..mars_adapters.core import MARS_DATAFRAME_CALLABLES

    module_methods: Dict[str, Any] = dict()
    with warnings.catch_warnings():
        # suppress warnings raised by pandas when import xorbits.pandas
        warning_members = [
            "pandas.Float64Index",
            "pandas.Int64Index",
            "pandas.UInt64Index",
        ]
        for m in warning_members:
            warning_message = (
                f"{m} is deprecated and will be removed from pandas in a future version. "
                "Use pandas.Index with the appropriate dtype instead."
            )
            warnings.filterwarnings(
                "ignore", category=FutureWarning, message=warning_message
            )
        for name, cls_member in inspect.getmembers(pd):
            if (
                name not in MARS_DATAFRAME_CALLABLES
                and inspect.isfunction(cls_member)
                and not name.startswith("_")
            ):
                warning_str = f"xorbits.pandas.{name} will fallback to Pandas"
                pandas_cls_method_wrapper = PandasClsMethodWrapper()
                output_type = pandas_cls_method_wrapper._get_output_type(
                    func=getattr(pd, name)
                )

                module_methods[name] = wrap_mars_callable(
                    wrap_fallback_module_method(pd, name, output_type, warning_str),
                    attach_docstring=True,
                    is_cls_member=False,
                    docstring_src_module=pd,
                    docstring_src=getattr(pd, name, None),
                    fallback_warning=True,
                )
    return module_methods



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional, Union

import xoscar as mo
from xoscar.serialization import AioDeserializer

from ...storage import StorageLevel, get_storage_backend
from ...storage.core import StorageFileObject
from ...typing import BandType
from ...utils import calc_data_size, lazy_import
from ..cluster import ClusterAPI, StorageInfo
from ..meta import MetaAPI
from .core import (
    DataInfo,
    DataManagerActor,
    StorageQuotaActor,
    WrappedStorageFileObject,
    build_data_info,
)
from .errors import DataNotExist, NoDataToSpill

cupy = lazy_import("cupy")
cudf = lazy_import("cudf")

logger = logging.getLogger(__name__)


class StorageHandlerActor(mo.Actor):
    """
    Storage handler actor, provide methods like `get`, `put`, etc.
    This actor is stateful and created on worker's sub pools.
    """

    def __init__(
        self,
        storage_init_params: Dict,
        data_manager_ref: mo.ActorRefType[DataManagerActor],
        spill_manager_refs,
        quota_refs: Dict[StorageLevel, mo.ActorRefType[StorageQuotaActor]],
        band_name: str = "numa-0",
    ):
        from .spill import SpillManagerActor

        self._storage_init_params = storage_init_params
        self._data_manager_ref = data_manager_ref
        self._spill_manager_refs: Dict[
            StorageLevel, mo.ActorRefType[SpillManagerActor]
        ] = spill_manager_refs
        self._quota_refs = quota_refs
        self._band_name = band_name
        self._supervisor_address = None

    @classmethod
    def gen_uid(cls, band_name: str):
        return f"storage_handler_{band_name}"

    @property
    def highest_level(self):
        return min(self._quota_refs)

    async def __post_create__(self):
        self._clients = clients = dict()
        for backend, init_params in self._storage_init_params.items():
            logger.debug("Start storage %s with params %s", backend, init_params)
            storage_cls = get_storage_backend(backend)
            client = storage_cls(**init_params)
            for level in StorageLevel.__members__.values():
                if client.level & level:
                    clients[level] = client

    def is_seekable(self, level: StorageLevel):
        if level is None:
            level = self.highest_level
        return self._clients[level].is_seekable

    async def _get_data(self, data_info: DataInfo, conditions: List[Any]):
        if data_info.offset is not None:
            reader = await self._clients[data_info.level].open_reader(
                data_info.object_id
            )
            await reader.seek(data_info.offset)
            res = await AioDeserializer(reader).run()
            if conditions is not None:
                try:
                    res = res.iloc[tuple(conditions)]
                except AttributeError:  # pragma: no cover
                    res = res[tuple(conditions)]
        elif conditions is None:
            res = yield self._clients[data_info.level].get(data_info.object_id)
        else:
            try:
                res = yield self._clients[data_info.level].get(
                    data_info.object_id, conditions=conditions
                )
            except NotImplementedError:
                data = yield self._clients[data_info.level].get(data_info.object_id)

                try:
                    from ...dataframe.utils import is_cudf

                    if is_cudf(data) and len(conditions) == 1:
                        sliced_value = data.iloc[conditions[0]]
                    else:
                        sliced_value = data.iloc[tuple(conditions)]
                except AttributeError:
                    sliced_value = data[tuple(conditions)]
                res = sliced_value
        raise mo.Return(res)

    @mo.extensible
    async def get(
        self,
        session_id: str,
        data_key: str,
        conditions: List = None,
        to_cpu: Optional[bool] = None,
        error: str = "raise",
    ):
        try:
            data_info = await self._data_manager_ref.get_data_info(
                session_id, data_key, self._band_name
            )
            data = yield self._get_data(data_info, conditions)

            from ...dataframe.utils import is_cudf
            from ...tensor.array_utils import is_cupy

            if (is_cudf(data) or is_cupy(data)) and to_cpu is True:
                data = data.to_pandas() if is_cudf(data) else data.get()
            raise mo.Return(data)
        except DataNotExist:
            if error == "raise":
                raise

    def _get_data_info(
        self,
        session_id: str,
        data_key: str,
        conditions: List = None,
        to_cpu: Optional[bool] = None,
        error: str = "raise",
    ):
        info = self._data_manager_ref.get_data_info.delay(
            session_id, data_key, self._band_name, error
        )
        return info, conditions, to_cpu

    @get.batch
    async def batch_get(self, args_list, kwargs_list):
        infos = []
        conditions_list = []
        to_cpus = []
        for args, kwargs in zip(args_list, kwargs_list):
            info, conditions, to_cpu = self._get_data_info(*args, **kwargs)
            infos.append(info)
            conditions_list.append(conditions)
            to_cpus.append(to_cpu)
        data_infos = await self._data_manager_ref.get_data_info.batch(*infos)
        results = []
        writer_args = [
            (info.object_id, info.level)
            for info in data_infos
            if info is not None and info.offset is not None
        ]
        object_id_to_reader = dict()
        for object_id, level in writer_args:
            object_id_to_reader[object_id] = await self._clients[level].open_reader(
                object_id
            )
        for data_info, conditions, to_cpu in zip(data_infos, conditions_list, to_cpus):
            if data_info is None:  # pragma: no cover
                result = None
            elif data_info.offset is not None:
                reader = object_id_to_reader[data_info.object_id]
                await reader.seek(data_info.offset)
                result = await AioDeserializer(reader).run()
            else:
                result = yield self._get_data(data_info, conditions)

            from ...dataframe.utils import is_cudf
            from ...tensor.array_utils import is_cupy

            if (is_cudf(result) or is_cupy(result)) and to_cpu is True:
                result = result.to_pandas() if is_cudf(result) else result.get()
            results.append(result)
        raise mo.Return(results)

    def _get_default_level(self, obj):
        from ...dataframe.utils import get_storage_level_gpu_or_memory

        if self.highest_level != StorageLevel.GPU:
            return self.highest_level
        else:
            return get_storage_level_gpu_or_memory(obj)

    @mo.extensible
    async def put(
        self, session_id: str, data_key: str, obj: object, level: StorageLevel = None
    ) -> DataInfo:
        if level is None:
            level = self._get_default_level(obj)
        size = await asyncio.to_thread(calc_data_size, obj)
        await self.request_quota_with_spill(level, size)
        object_info = await self._clients[level].put(obj)
        data_info = build_data_info(object_info, level, size, self._band_name)
        await self._data_manager_ref.put_data_info(
            session_id, data_key, data_info, object_info
        )
        if object_info.size is not None and data_info.memory_size != object_info.size:
            await self._quota_refs[level].update_quota(
                object_info.size - data_info.memory_size
            )
        await self.notify_spillable_space(level)
        return data_info

    @put.batch
    async def batch_put(self, args_list, kwargs_list):
        objs = []
        data_keys = []
        session_id = None
        level = last_level = None
        sizes = []
        for args, kwargs in zip(args_list, kwargs_list):
            session_id, data_key, obj, level = self.put.bind(*args, **kwargs)
            if level is None:
                level = self._get_default_level(obj)
            size = await asyncio.to_thread(calc_data_size, obj)
            if last_level is not None:
                assert last_level == level
            last_level = level
            objs.append(obj)
            data_keys.append(data_key)
            sizes.append(size)

        await self.request_quota_with_spill(level, sum(sizes))

        data_infos = []
        put_infos = []
        quota_delta = 0
        for size, data_key, obj in zip(sizes, data_keys, objs):
            object_info = await self._clients[level].put(obj)
            data_info = build_data_info(object_info, level, size, self._band_name)
            data_infos.append(data_info)
            if (
                object_info.size is not None
                and data_info.memory_size != object_info.size
            ):
                # we request memory size before putting, when put finishes,
                # update quota to the true store size
                quota_delta += object_info.size - data_info.memory_size
            put_infos.append(
                self._data_manager_ref.put_data_info.delay(
                    session_id, data_key, data_info, object_info
                )
            )
        await self._quota_refs[level].update_quota(quota_delta)
        await self._data_manager_ref.put_data_info.batch(*put_infos)
        await self.notify_spillable_space(level)
        return data_infos

    async def delete_object(
        self,
        session_id: str,
        data_key: Any,
        data_size: Union[int, float],
        object_id: Any,
        level: StorageLevel,
    ):
        data_key = await self._data_manager_ref.get_store_key(session_id, data_key)
        await self._data_manager_ref.delete_data_info(
            session_id, data_key, level, self._band_name
        )
        await self._clients[level].delete(object_id)
        await self._quota_refs[level].release_quota(data_size)

    @mo.extensible
    async def delete(self, session_id: str, data_key: str, error: str = "raise"):
        if error not in ("raise", "ignore"):  # pragma: no cover
            raise ValueError("error must be raise or ignore")

        data_key = await self._data_manager_ref.get_store_key(session_id, data_key)
        if isinstance(data_key, list):
            # delete mapper main key
            data_keys = data_key
        else:
            data_keys = [data_key]
        for data_key in data_keys:
            all_infos = await self._data_manager_ref.get_data_infos(
                session_id, data_key, self._band_name, error
            )
            if not all_infos:
                return

            key_to_infos = (
                all_infos if isinstance(all_infos, dict) else {data_key: all_infos}
            )

            for key, infos in key_to_infos.items():
                for info in infos:
                    level = info.level
                    await self._data_manager_ref.delete_data_info(
                        session_id, key, level, self._band_name
                    )
                    await self._clients[level].delete(info.object_id)
                    await self._quota_refs[level].release_quota(info.store_size)

    @delete.batch
    async def batch_delete(self, args_list, kwargs_list):
        session_id = None
        error = None
        data_keys = []
        for args, kwargs in zip(args_list, kwargs_list):
            session_id, data_key, error = self.delete.bind(*args, **kwargs)
            data_keys.append(
                self._data_manager_ref.get_store_key.delay(session_id, data_key)
            )
        store_keys = await self._data_manager_ref.get_store_key.batch(*data_keys)
        data_keys = set()
        for k in store_keys:
            if isinstance(k, list):
                data_keys.update(set(k))
            else:
                data_keys.add(k)

        infos_list = await self._data_manager_ref.get_data_infos.batch(
            *[
                self._data_manager_ref.get_data_infos.delay(
                    session_id, data_key, self._band_name, error
                )
                for data_key in data_keys
            ]
        )

        delete_infos = []
        to_removes = []
        level_sizes = defaultdict(lambda: 0)
        for all_infos, data_key in zip(infos_list, data_keys):
            if not all_infos:
                # data not exist and error == 'ignore'
                continue
            key_to_infos = (
                all_infos if isinstance(all_infos, dict) else {data_key: all_infos}
            )

            for key, infos in key_to_infos.items():
                for info in infos:
                    level = info.level
                    delete_infos.append(
                        self._data_manager_ref.delete_data_info.delay(
                            session_id, key, level, info.band
                        )
                    )
                    to_removes.append((level, info.object_id))
                    level_sizes[level] += info.store_size

        if not delete_infos:
            # no data to remove
            return

        await self._data_manager_ref.delete_data_info.batch(*delete_infos)
        await asyncio.gather(
            *[self._clients[level].delete(object_id) for level, object_id in to_removes]
        )
        for level, size in level_sizes.items():
            await self._quota_refs[level].release_quota(size)

    @mo.extensible
    async def open_reader(self, session_id: str, data_key: str) -> StorageFileObject:
        data_info = await self._data_manager_ref.get_data_info(
            session_id, data_key, self._band_name
        )
        reader = await self._clients[data_info.level].open_reader(data_info.object_id)
        return reader

    @open_reader.batch
    async def batch_open_readers(self, args_list, kwargs_list):
        get_data_infos = []
        for args, kwargs in zip(args_list, kwargs_list):
            get_data_infos.append(
                self._data_manager_ref.get_data_info.delay(
                    *args, band_name=self._band_name, **kwargs
                )
            )
        data_infos = await self._data_manager_ref.get_data_info.batch(*get_data_infos)
        return await asyncio.gather(
            *[
                self._clients[data_info.level].open_reader(data_info.object_id)
                for data_info in data_infos
            ]
        )

    @mo.extensible
    async def open_writer(
        self,
        session_id: str,
        data_key: str,
        size: int,
        level: StorageLevel,
        request_quota=True,
        band_name: str = "numa-0",
    ) -> WrappedStorageFileObject:
        if level is None:
            level = self.highest_level
        if request_quota:
            await self.request_quota_with_spill(level, size)
        writer = await self._clients[level].open_writer(size)
        return WrappedStorageFileObject(
            writer,
            level,
            size,
            session_id,
            data_key,
            self._data_manager_ref,
            self._clients[level],
            band_name,
        )

    @open_writer.batch
    async def batch_open_writers(self, args_list, kwargs_list):
        extracted_args = None
        data_keys, sizes = [], []
        for args, kwargs in zip(args_list, kwargs_list):
            (
                session_id,
                data_key,
                size,
                level,
                request_quota,
                band_name,
            ) = self.open_writer.bind(*args, **kwargs)
            if extracted_args:
                assert extracted_args == (session_id, level, request_quota, band_name)
            extracted_args = (session_id, level, request_quota, band_name)
            data_keys.append(data_key)
            sizes.append(size)
        session_id, level, request_quota, band_name = extracted_args
        if level is None:  # pragma: no cover
            level = self.highest_level
        if request_quota:  # pragma: no cover
            await self.request_quota_with_spill(level, sum(sizes))
        writers = await asyncio.gather(
            *[self._clients[level].open_writer(size) for size in sizes]
        )
        wrapped_writers = []
        for writer, size, data_key in zip(writers, sizes, data_keys):
            wrapped_writers.append(
                WrappedStorageFileObject(
                    writer,
                    level,
                    size,
                    session_id,
                    data_key,
                    self._data_manager_ref,
                    self._clients[level],
                    band_name=band_name,
                )
            )
        return wrapped_writers

    async def _get_meta_api(self, session_id: str):
        if self._supervisor_address is None:
            cluster_api = await ClusterAPI.create(self.address)
            [self._supervisor_address] = await cluster_api.get_supervisors_by_keys(
                [session_id]
            )

        return await MetaAPI.create(
            session_id=session_id, address=self._supervisor_address
        )

    async def _fetch_remote(
        self,
        session_id: str,
        data_keys: List[Union[str, tuple]],
        remote_band: BandType,
        error: str,
    ):
        remote_manager_ref: mo.ActorRefType[DataManagerActor] = await mo.actor_ref(
            uid=DataManagerActor.default_uid(), address=remote_band[0]
        )
        get_data_infos = []
        for data_key in data_keys:
            get_data_infos.append(
                remote_manager_ref.get_data_info.delay(session_id, data_key, error)
            )
        data_infos = await remote_manager_ref.get_data_info.batch(*get_data_infos)
        data_infos, data_keys = zip(
            *[
                (data_info, data_key)
                for data_info, data_key in zip(data_infos, data_keys)
                if data_info is not None
            ]
        )
        put_data_info_delays = []
        fetch_tasks = []
        for data_info, data_key in zip(data_infos, data_keys):
            put_data_info_delays.append(
                self._data_manager_ref.put_data_info.delay(
                    session_id, data_key, data_info, None
                )
            )
            fetch_tasks.append(
                self._clients[StorageLevel.REMOTE].fetch(data_info.object_id)
            )
        await self._data_manager_ref.put_data_info.batch(*put_data_info_delays)
        await asyncio.gather(*fetch_tasks)

    async def _fetch_via_transfer(
        self,
        session_id: str,
        data_keys: List[Union[str, tuple]],
        level: StorageLevel,
        remote_band: BandType,
        fetch_band_name: str,
        error: str,
    ):
        from .transfer import SenderManagerActor

        logger.debug("Begin to fetch %s from band %s", data_keys, remote_band)
        sender_ref: mo.ActorRefType[SenderManagerActor] = await mo.actor_ref(
            address=remote_band[0], uid=SenderManagerActor.gen_uid(remote_band[1])
        )
        await sender_ref.send_batch_data(
            session_id,
            data_keys,
            self._data_manager_ref.address,
            level,
            fetch_band_name,
            error=error,
        )
        logger.debug("Finish fetching %s from band %s", data_keys, remote_band)

    async def fetch_batch(
        self,
        session_id: str,
        data_keys: List[str],
        level: StorageLevel,
        band_name: str,
        address: str,
        error: str,
    ):
        if error not in ("raise", "ignore"):  # pragma: no cover
            raise ValueError("error must be raise or ignore")

        meta_api = await self._get_meta_api(session_id)
        remote_keys = defaultdict(set)
        missing_keys = []
        get_metas = []
        get_info_delays = []
        for data_key in data_keys:
            get_info_delays.append(
                self._data_manager_ref.get_data_info.delay(
                    session_id, data_key, band_name, error="ignore"
                )
            )
        data_infos = await self._data_manager_ref.get_data_info.batch(*get_info_delays)
        pin_delays = []
        for data_key, info in zip(data_keys, data_infos):
            # for gpu bands, need transfer between gpu cards
            if info is not None:
                if band_name and band_name != info.band:
                    missing_keys.append(data_key)
                else:
                    pin_delays.append(
                        self._data_manager_ref.pin.delay(
                            session_id, data_key, self._band_name
                        )
                    )
            else:
                # Not exists in local, fetch from remote worker
                missing_keys.append(data_key)
        if address is None or band_name is None:
            # some mapper keys are absent, specify error='ignore'
            # remember that meta only records those main keys
            get_metas = [
                (
                    meta_api.get_chunk_meta.delay(
                        data_key[0] if isinstance(data_key, tuple) else data_key,
                        fields=["bands"],
                        error="ignore",
                    )
                )
                for data_key in missing_keys
            ]
        await self._data_manager_ref.pin.batch(*pin_delays)

        if get_metas:
            metas = await meta_api.get_chunk_meta.batch(*get_metas)
        else:  # pragma: no cover
            metas = [{"bands": [(address, band_name)]}] * len(missing_keys)
        assert len(metas) == len(missing_keys)
        for data_key, bands in zip(missing_keys, metas):
            if bands is not None:
                remote_keys[bands["bands"][0]].add(data_key)
        transfer_tasks = []
        fetch_keys = []
        for band, keys in remote_keys.items():
            if StorageLevel.REMOTE in self._quota_refs:
                # if storage support remote level, just fetch object id
                transfer_tasks.append(
                    self._fetch_remote(session_id, list(keys), band, error)
                )
            else:
                # fetch via transfer
                transfer_tasks.append(
                    self._fetch_via_transfer(
                        session_id, list(keys), level, band, band_name or band[1], error
                    )
                )
            fetch_keys.extend(list(keys))

        await asyncio.gather(*transfer_tasks)

        set_meta_keys = set()
        for data_key in fetch_keys:
            # skip shuffle keys
            if isinstance(data_key, tuple):
                set_meta_keys.add(data_key[0])
            else:
                set_meta_keys.add(data_key)
        append_bands_delays = [
            meta_api.add_chunk_bands.delay(key, [(self.address, self._band_name)])
            for key in set_meta_keys
        ]

        if append_bands_delays:
            await meta_api.add_chunk_bands.batch(*append_bands_delays)

    async def request_quota_with_spill(self, level: StorageLevel, size: int):
        if await self._quota_refs[level].request_quota(size):
            return
        else:
            total, used = await self._quota_refs[level].get_quota()
            await self.spill(level, int(used + size - total), size)
            await self._quota_refs[level].request_quota(size)
            logger.debug(
                "Spill is triggered, request %s bytes of %s finished", size, level
            )

    async def notify_spillable_space(self, level):
        if await self._spill_manager_refs[level].has_spill_task():
            total, used = await self._quota_refs[level].get_quota()
            tasks = []
            if total is not None:
                spillable_size = await self._data_manager_ref.get_spillable_size(
                    level, self._band_name
                )
                tasks.append(
                    self._spill_manager_refs[level].notify_spillable_space(
                        spillable_size, total - used
                    )
                )
            await asyncio.gather(*tasks)

    async def spill(self, level: StorageLevel, request_size: int, object_size: int):
        from .spill import spill

        try:
            await spill(
                request_size, level, self._band_name, self._data_manager_ref, self
            )
        except NoDataToSpill:
            logger.warning(
                "No data to spill %s bytes, waiting more space", request_size
            )
            size = await self._spill_manager_refs[level].wait_for_space(object_size)
            await spill(size, level, self._band_name, self._data_manager_ref, self)

    async def list(self, level: StorageLevel) -> List:
        return await self._data_manager_ref.list(level, self._band_name)

    @mo.extensible
    async def unpin(self, session_id: str, data_key: str, error: str = "raise"):
        levels = await self._data_manager_ref.unpin(
            session_id, [data_key], self._band_name, error
        )
        if levels:
            await self.notify_spillable_space(levels[0])

    @unpin.batch
    async def batch_unpin(self, args_list, kwargs_list):
        extracted_args = []
        data_keys = []
        for args, kw in zip(args_list, kwargs_list):
            session_id, data_key, error = self.unpin.bind(*args, **kw)
            if extracted_args:
                assert extracted_args == (session_id, error)
            extracted_args = session_id, error
            data_keys.append(data_key)
        if extracted_args:
            session_id, error = extracted_args
            levels = await self._data_manager_ref.unpin(
                session_id, data_keys, self._band_name, error
            )
            for level in levels:
                await self.notify_spillable_space(level)

    async def get_storage_level_info(self, level: StorageLevel) -> StorageInfo:
        quota_ref = self._quota_refs[level]
        total_size, used_size = await quota_ref.get_quota()
        return StorageInfo(
            storage_level=level,
            total_size=int(total_size) if total_size else total_size,
            used_size=int(used_size),
        )

    async def get_storage_backend_info(self, level: StorageLevel) -> dict:
        return self._clients[level].backend_info



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from typing import Callable, Dict, List, Optional

from ..core.adapter import to_mars, wrap_mars_callable


class SKLearnBase:
    def __init__(self, *args, **kwargs):
        self.mars_instance = self._marscls(*to_mars(args), **to_mars(kwargs))

    def __getattr__(self, name):
        return getattr(self.mars_instance, name)


def wrap_cls_func(marscls: Callable, name: str, submodule):
    @functools.wraps(getattr(marscls, name))
    def wrapped(self, *args, **kwargs):
        return getattr(self.mars_instance, name)(*args, **kwargs)

    return wrap_mars_callable(
        wrapped,
        member_name=name,
        attach_docstring=True,
        is_cls_member=True,
        docstring_src_module=submodule,
        docstring_src_cls=getattr(submodule, marscls.__name__, None),
    )


def _collect_module_callables(
    mars_module,
    orig_module,
    skip_members: Optional[List[str]] = None,
) -> Dict[str, Callable]:
    module_callables: Dict[str, Callable] = dict()

    for name, func in inspect.getmembers(mars_module, inspect.isfunction):
        if skip_members is not None and name in skip_members:
            continue
        module_callables[name] = wrap_mars_callable(
            func,
            attach_docstring=True,
            is_cls_member=False,
            docstring_src_module=orig_module,
            docstring_src=getattr(orig_module, name, None),
        )
    return module_callables


def _install_cls_members(
    module_cls_map, module_callables: Dict[str, Callable], orig_submodule
):
    for x_cls, mars_cls in module_cls_map.items():
        module_callables[x_cls.__name__] = x_cls
        for name, _ in inspect.getmembers(mars_cls, inspect.isfunction):
            if not name.startswith("_"):
                setattr(x_cls, name, wrap_cls_func(mars_cls, name, orig_submodule))



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import xgboost
except ImportError:  # pragma: no cover
    xgboost = None

MARS_XGBOOST_CALLABLES = {}

if xgboost is not None:
    import functools
    import inspect
    from typing import Callable, Dict, List, Optional

    from ..._mars.learn.contrib import xgboost as mars_xgboost
    from ..._mars.learn.contrib.xgboost.classifier import (
        XGBClassifier as MarsXGBClassifier,
    )
    from ..._mars.learn.contrib.xgboost.regressor import (
        XGBRegressor as MarsXGBRegressor,
    )
    from ...core.adapter import to_mars, wrap_mars_callable
    from ...core.utils.docstring import attach_module_callable_docstring

    class BaseXGB:
        def __init__(self, *args, **kwargs):
            self.mars_instance = self._mars_cls(*to_mars(args), **to_mars(kwargs))

    class XGBClassifier(BaseXGB):
        _mars_cls = MarsXGBClassifier

    class XGBRegressor(BaseXGB):
        _mars_cls = MarsXGBRegressor

    xgboost_class_mappings: Dict = {
        XGBClassifier: MarsXGBClassifier,
        XGBRegressor: MarsXGBRegressor,
    }

    def wrap_cls_func(mars_cls: Callable, name: str):
        @functools.wraps(getattr(mars_cls, name))
        def wrapped(self, *args, **kwargs):
            return getattr(self.mars_instance, name)(*args, **kwargs)

        return wrap_mars_callable(
            wrapped,
            member_name=name,
            attach_docstring=True,
            is_cls_member=True,
            docstring_src_module=xgboost,
            docstring_src_cls=getattr(xgboost, mars_cls.__name__, None),
        )

    def _collect_module_callables(
        skip_members: Optional[List[str]] = None,
    ) -> Dict[str, Callable]:
        module_callables: Dict[str, Callable] = dict()

        for name, func in inspect.getmembers(mars_xgboost, inspect.isfunction):
            if skip_members is not None and name in skip_members:
                continue

            if name == "MarsDMatrix":
                name = "DMatrix"

            module_callables[name] = wrap_mars_callable(
                func,
                attach_docstring=True,
                is_cls_member=False,
                docstring_src_module=xgboost,
                docstring_src=getattr(xgboost, name, None),
            )
        return module_callables

    def _install_cls_members(module_callables: Dict[str, Callable]):
        for k, v in xgboost_class_mappings.items():
            module_callables[k.__name__] = k
            for name, _ in inspect.getmembers(v, inspect.isfunction):
                if not name.startswith("_"):
                    setattr(k, name, wrap_cls_func(v, name))

    MARS_XGBOOST_CALLABLES = _collect_module_callables(skip_members=["register_op"])

    _install_cls_members(MARS_XGBOOST_CALLABLES)

    attach_module_callable_docstring(XGBClassifier, xgboost, xgboost.XGBClassifier)
    attach_module_callable_docstring(XGBRegressor, xgboost, xgboost.XGBRegressor)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...serialization.serializables import (
    AnyField,
    BoolField,
    DictField,
    Int32Field,
    Int64Field,
    KeyField,
    Serializable,
    StringField,
    TupleField,
)
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index, validate_axis

_PAIRWISE_AGG = ["corr", "cov"]


class GroupByRolling(Serializable):
    input = KeyField("input")
    window = AnyField("window")
    min_periods = Int64Field("min_periods")
    center = BoolField("center")
    win_type = StringField("win_type")
    on = StringField("on")
    axis = Int32Field("axis")
    closed = StringField("closed")

    def validate(self):
        """
        leverage pandas itself to validate the parameters.
        """
        groupby = self.input
        groupby.op.build_mock_groupby().rolling(**self.params)

    @property
    def params(self):
        p = OrderedDict()
        args = [
            "window",
            "min_periods",
            "center",
            "win_type",
            "axis",
            "on",
            "closed",
        ]
        for attr in args:
            p[attr] = getattr(self, attr)
        return p

    def aggregate(self, func, *args, **kwargs):
        params = self.params
        if func in _PAIRWISE_AGG:
            # for convenience, since pairwise aggregations are axis irrelevant.
            params["axis"] = 0

        op = GroupbyRollingAgg(func=func, func_args=args, func_kwargs=kwargs, **params)
        return op(self)

    def agg(self, func, *args, **kwargs):
        return self.aggregate(func, *args, **kwargs)

    def count(self):
        return self.aggregate("count")

    def sum(self, *args, **kwargs):
        return self.aggregate("sum", *args, **kwargs)

    def mean(self, *args, **kwargs):
        return self.aggregate("mean", *args, **kwargs)

    def median(self, **kwargs):
        return self.aggregate("median", **kwargs)

    def var(self, ddof=1, *args, **kwargs):
        return self.aggregate("var", ddof=ddof, *args, **kwargs)

    def std(self, ddof=1, *args, **kwargs):
        return self.aggregate("std", ddof=ddof, *args, **kwargs)

    def min(self, *args, **kwargs):
        return self.aggregate("min", *args, **kwargs)

    def max(self, *args, **kwargs):
        return self.aggregate("max", *args, **kwargs)

    def skew(self, **kwargs):
        return self.aggregate("skew", **kwargs)

    def kurt(self, **kwargs):
        return self.aggregate("kurt", **kwargs)

    def corr(self, **kwargs):
        return self.aggregate("corr", **kwargs)

    def cov(self, **kwargs):
        return self.aggregate("cov", **kwargs)


class GroupbyRollingAgg(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.GROUPBY_ROLLING_AGG

    input = KeyField("input")
    window = AnyField("window")
    min_periods = Int64Field("min_periods")
    center = BoolField("center")
    win_type = StringField("win_type")
    on = StringField("on")
    axis = Int32Field("axis")
    closed = StringField("closed")
    func = AnyField("func")
    func_args = TupleField("func_args")
    func_kwargs = DictField("func_kwargs")

    def __call__(self: "GroupbyRollingAgg", r: "GroupByRolling"):
        groupby = r.input
        mock_obj = (
            groupby.op.build_mock_groupby()
            .rolling(**r.params)
            .agg(self.func, *self.func_args, **self.func_kwargs)
        )

        index_value = parse_index(mock_obj.index)
        if isinstance(mock_obj, pd.DataFrame):
            return self.new_dataframe(
                [groupby],
                shape=(np.nan, mock_obj.shape[1]),
                dtypes=mock_obj.dtypes,
                index_value=index_value,
                columns_value=parse_index(mock_obj.dtypes.index, store_data=True),
            )
        elif isinstance(mock_obj, pd.Series):
            return self.new_series(
                [groupby],
                shape=(np.nan,),
                dtype=mock_obj.dtype,
                index_value=index_value,
                name=mock_obj.name,
            )

    @classmethod
    def tile(cls, op: "GroupbyRollingAgg"):
        in_groupby = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_groupby.chunks:
            inp_chunks = [c]

            new_op = op.copy().reset_key()
            new_op.tileable_op_key = op.key
            if op.output_types[0] == OutputType.dataframe:
                chunks.append(
                    new_op.new_chunk(
                        inp_chunks,
                        index=c.index,
                        shape=(np.nan, len(out_df.dtypes)),
                        dtypes=out_df.dtypes,
                        columns_value=out_df.columns_value,
                        index_value=out_df.index_value,
                    )
                )
            else:
                chunks.append(
                    new_op.new_chunk(
                        inp_chunks,
                        name=out_df.name,
                        index=(c.index[0],),
                        shape=(np.nan,),
                        dtype=out_df.dtype,
                        index_value=out_df.index_value,
                    )
                )

        new_op = op.copy()
        kw = out_df.params.copy()
        kw["chunks"] = chunks
        if op.output_types[0] == OutputType.dataframe:
            kw["nsplits"] = ((np.nan,) * len(chunks), (out_df.shape[1],))
        else:
            kw["nsplits"] = ((np.nan,) * len(chunks),)
        return new_op.new_tileable([in_groupby], **kw)

    @classmethod
    def execute(cls, ctx, op: "GroupbyRollingAgg"):
        inp = ctx[op.inputs[0].key]

        r = inp.rolling(
            window=op.window,
            min_periods=op.min_periods,
            center=op.center,
            win_type=op.win_type,
            on=op.on,
            axis=op.axis,
            closed=op.closed,
        )
        result = r.aggregate(op.func, *op.func_args, **op.func_kwargs)

        ctx[op.outputs[0].key] = result


def rolling(
    groupby,
    window,
    min_periods=None,
    center=False,
    win_type=None,
    on=None,
    axis=0,
    closed=None,
) -> GroupByRolling:
    axis = validate_axis(axis, groupby)
    r = GroupByRolling(
        input=groupby,
        window=window,
        min_periods=min_periods,
        center=center,
        win_type=win_type,
        on=on,
        axis=axis,
        closed=closed,
    )
    r.validate()

    return r



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy

from ...core.adapter import (
    MarsCClass,
    MarsGetItemProxy,
    MarsMGridClass,
    MarsOGridClass,
    MarsRClass,
    register_converter,
)
from ...core.utils.docstring import attach_module_callable_docstring


@register_converter(from_cls_list=[MarsCClass])
class CClass(MarsGetItemProxy):
    def __init__(self):
        super().__init__(MarsCClass())


attach_module_callable_docstring(CClass, numpy, numpy.lib.index_tricks.CClass)
c_ = CClass()


@register_converter(from_cls_list=[MarsRClass])
class RClass(MarsGetItemProxy):
    def __init__(self):
        super().__init__(MarsRClass())


attach_module_callable_docstring(RClass, numpy, numpy.lib.index_tricks.RClass)
r_ = RClass()


@register_converter(from_cls_list=[MarsOGridClass])
class OGridClass(MarsGetItemProxy):
    def __init__(self):
        super().__init__(MarsOGridClass())


attach_module_callable_docstring(OGridClass, numpy, numpy.lib.index_tricks.OGridClass)
ogrid = OGridClass()


@register_converter(from_cls_list=[MarsMGridClass])
class MGridClass(MarsGetItemProxy):
    def __init__(self):
        super().__init__(MarsMGridClass())


attach_module_callable_docstring(MGridClass, numpy, numpy.lib.index_tricks.MGridClass)
mgrid = MGridClass()



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.semi_supervised as sk_ss

from ...._mars.learn import semi_supervised as mars_ss
from ...._mars.learn.semi_supervised import LabelPropagation as MarsLabelPropagation
from ....core.utils.docstring import attach_module_callable_docstring
from ...utils import SKLearnBase, _collect_module_callables, _install_cls_members


class LabelPropagation(SKLearnBase):
    _marscls = MarsLabelPropagation


SKLEARN_SS_CLS_MAP = {
    LabelPropagation: MarsLabelPropagation,
}

MARS_SKLEARN_SS_CALLABLES = _collect_module_callables(
    mars_ss, sk_ss, skip_members=["register_op"]
)
_install_cls_members(SKLEARN_SS_CLS_MAP, MARS_SKLEARN_SS_CALLABLES, sk_ss)
attach_module_callable_docstring(LabelPropagation, sk_ss, sk_ss.LabelPropagation)
